from enum import Enum

class CDCadentialStates(Enum):
    Idle = 1
    CadExpected = 2
    CadInevitable = 3
    PACArrival = 4
    CadAvoided = 5
    IACArrival = 6
    HCArrival = 7
    HCArrivalExpected = 8
    PCCArrival = 9
    PACAppoggExpected = 10
    IACAppoggExpected = 11
    IACArrivalExpected = 12
    HCAppoggExpected = 13
    BassAppoggExpected = 14
    PCHArrival = 15

class CDHarmonicChordDegrees(Enum):
    I = 1
    II = 2
    III = 3
    IV = 4
    V = 5
    VI = 6
    VII = 7


class CDHarmonicChordInversions(Enum):
    Root = 0
    First = 1
    Second = 2
    Third = 3

class CDKeyDetectionModes(Enum):
    FromFile = 0
    KSRaw = 1
    KSWithSmoothing = 2
    KSWithSmoothingCadenceSensitive = 3

class CDHarmonicState:
    def __init__(self, Key, Chord, ChordWithRests, ChordDegree, ChordInversion, ChordFigure, Alberti, Arpeggio, TimeSig, RomanNumeral, RealNotes):
        self.Key = Key
        self.Chord = Chord
        self.RomanNumeral = RomanNumeral
        self.ChordWithBassRests = ChordWithRests
        self.ChordDegree = ChordDegree
        self.ChordInversion = ChordInversion
        self.ChordFigure = ChordFigure
        self.Alberti = Alberti
        self.Arpeggio = Arpeggio
        self.TimeSig = TimeSig
        self.RealNotes = RealNotes
        # caching variables
        self.bassPitchFavorByPart = [None, None]
        self.sopranoPitch = None
        self.isDominant = None
        self.isTonic = None
        self.isBassRest = None


SUB = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")
SUP = str.maketrans("0123456789", "⁰¹²³⁴⁵⁶⁷⁸⁹")



from .CadenceDetectStateMachine import *
import music21 as m21
import math
import os
import pickle
import pickle
import enum
import copy
import sys

class CadenceDetector:
    def __init__(self,
                 maxNumMeasures=500,
                 minInitialMeasures=0,
                 minPostCadenceMeasures=0,
                 keyDetectionMode = CDKeyDetectionModes.KSWithSmoothingCadenceSensitive,
                 keyDetectionForgetFactor = 0.8,
                 reenforcementFactors = {'PAC': 3, 'IAC': 1, 'HC': 3/2},
                 keyDetectionLookahead = 0.5,
                 keyDetectionBlockSize=4,
                 keyDetectionOverlap=0.25):
        self.MaxNumMeasures = maxNumMeasures
        self.KeyDetectionMode = keyDetectionMode
        self.KeyDetectionForgetFactor = keyDetectionForgetFactor
        self.ReenforcementFactors = reenforcementFactors
        self.key_detection_lookahead = keyDetectionLookahead
        self.blockSize = keyDetectionBlockSize
        self.overlap = keyDetectionOverlap
        self.HarmonicStateMachine = CDStateMachine(isChallenger=False, minInitialMeasures=minInitialMeasures, minPostCadenceMeasures=minPostCadenceMeasures)
        self.HarmonicStateMachineChallenger = CDStateMachine(isChallenger=True, minInitialMeasures=minInitialMeasures, minPostCadenceMeasures=minPostCadenceMeasures)
        self.NoteStream = []
        self.NoteStreamRestless = []
        self.NoteStreamReduced = []
        self.ChordStream = []
        self.ChordStreamRestless = []
        self.BassChords = []
        self.KeyPerMeasure = []
        self.CorrectedKeyPerMeasure = []
        self.InterpretationsPerMeasure = []
        self.SmoothedCorrCoefsPerMeasure = []
        self.CurrSmoothedInterpretations = {}
        #{'A': 0, 'B-': 0, 'B': 0, 'C': 0, 'C#': 0, 'D': 0,
        #'E-': 0, 'E': 0, 'F': 0, 'F#': 0, 'G': 0, 'A-': 0,
        #'a': 0, 'b-': 0, 'b': 0, 'c': 0, 'c#': 0, 'd': 0,
        #'e-': 0, 'e': 0, 'f': 0, 'f#': 0, 'g': 0, 'a-': 0}
        self.Top2Keys = []
        self.PrevKeyString = []
        self.PrevChallengerKeyString = []
        self.OptionalKeysPerMeasure = [[]]
        self.Parts = []
        self.MeasuresPerPart = []
        self.NumMeasures = 0
        self.fileName = 0
        self.hasPickupMeasure = 0
        self.WritePath = 0
        self.RepeatedMeasureByVolta = []
        self.IncompleteMeasures = []
        self.EmptyMeasures = []
        self.FinalBarlines = []
        self.analysis_full = []
        self.analysis_filtered = []
        self.analysis_transition_strings = []
        self.PrevMeasureAlberti = False
        self.PrevMeasureArpeggio = False


    def loadMusic21Corpus(self,fileString):
        self.NoteStream = m21.corpus.parse(fileString)
        self.analyze()

    def loadFileAndGetMeasures(self, fileString):
        print("Loading score...")
        self.NoteStream = m21.converter.parse(fileString)
        self.Parts = self.NoteStream.recurse().getElementsByClass(m21.stream.Part)
        self.NumParts = len(self.Parts)
        for part in self.Parts:
            self.MeasuresPerPart.append(part.recurse().getElementsByClass(m21.stream.Measure))
        self.NumMeasures = max([len(curr_part_measures) for curr_part_measures in self.MeasuresPerPart])
        self.NumMeasures = min(self.NumMeasures, self.MaxNumMeasures)
        #update some data to state machines
        self.HarmonicStateMachine.NumParts = self.NumParts
        self.HarmonicStateMachineChallenger.NumParts = self.NumParts
        # do some other quick analysis for debugging without full cadence detection
        # self.findVoltas()

    def loadFile(self,fileString):
        self.loadFileAndGetMeasures(fileString)
        self.analyze()

    def analyze(self):
        self.removeChordSymbolsInNoteStream()
        self.removePickupMeasure()
        self.addMyLablesToParts(self.NoteStream)
        NoteStreamNoGrace = self.removeGraceNotes(self.NoteStream)
        try:
            self.ChordStream = NoteStreamNoGrace.chordify(addPartIdAsGroup=True, removeRedundantPitches=False, copyPitches=False)
            self.HarmonicStateMachine.CheckBassPartFromChord = True
            self.HarmonicStateMachineChallenger.CheckBassPartFromChord = True
        except:
            print("Cannot add parts to chords!!")
            self.ChordStream = NoteStreamNoGrace.chordify(removeRedundantPitches=False, copyPitches=False)
            self.HarmonicStateMachine.CheckBassPartFromChord = False
            self.HarmonicStateMachineChallenger.CheckBassPartFromChord = False


        self.NumMeasures = len(self.ChordStream.recurse().getElementsByClass(m21.stream.Measure))
        self.NumMeasures = min(self.NumMeasures, self.MaxNumMeasures)
        self.findVoltas()
        self.findFinalBarlines()
        self.findEmptyMeasures()
        self.findIncompleteMeasures()
        self.NoteStreamRestless = copy.deepcopy(self.NoteStream)  #create new list so as not to alter the original stream
        self.replaceBassRestsWithPrevsViaSecondsMaps()
        self.addMyLablesToParts(self.NoteStreamRestless)
        NoteStreamRestlessNoGrace = self.removeGraceNotes(self.NoteStreamRestless)
        if self.HarmonicStateMachine.CheckBassPartFromChord==True:
            self.ChordStreamRestless = NoteStreamRestlessNoGrace.chordify(addPartIdAsGroup=True, removeRedundantPitches=False, copyPitches=False)
        else:
            self.ChordStreamRestless = NoteStreamRestlessNoGrace.chordify(removeRedundantPitches=False, copyPitches=False)

        #chordifying only bass part
        parts = self.NoteStream.recurse().getElementsByClass(m21.stream.Part)
        BassPartNoGrace = self.removeGraceNotes(parts[-1])
        try:
            self.BassChords = BassPartNoGrace.chordify(addPartIdAsGroup=True, removeRedundantPitches=False, copyPitches=False)
        except:
            self.BassChords = BassPartNoGrace.chordify(removeRedundantPitches=False, copyPitches=False)

        self.findGroupLimitsViaSecondsMap()

    def findGroupLimitsViaSecondsMap(self):
        for part in self.Parts:
            prev_secondsDict = prev_note = prev_diff = []
            part_measures_secondsMap = sorted(part.recurse().getElementsByClass(m21.stream.Measure).secondsMap, key=lambda d: d['offsetSeconds'])
            part_notes_secondsMap = sorted(part.recurse().notesAndRests.secondsMap, key=lambda d: d['offsetSeconds'])
            for secondsDict in part_notes_secondsMap:
                curr_note = secondsDict['element']
                if curr_note.tie and not curr_note.tie == m21.tie.Tie('start'):
                    continue
                if prev_secondsDict:
                    curr_diff = secondsDict['offsetSeconds'] - prev_secondsDict['offsetSeconds']
                    if curr_diff <= 0:
                        continue
                    prev_note = prev_secondsDict['element']
                    if (prev_note.isRest and curr_note.isRest): # no grouping on consecutive rests
                        pass
                    elif not prev_note.isRest and curr_note.isRest:
                        prev_note.groups.append('}')
                    elif prev_note.isRest and not curr_note.isRest:
                        curr_note.groups.append('{')
                    elif prev_diff:
                        # long note groups backward with previous shorter notes
                        # also close grouping on very long notes (longer than a measure)
                        # epsilon and frac needed because they may be different due to fraction and measure duration inaccuracy
                        epsilon = 0.001
                        frac = 0.9
                        # === debug per measure
                        if curr_note.measureNumber and curr_note.measureNumber==66:
                            bla = 0
                        if (curr_diff > prev_diff + epsilon) or (prev_note.measureNumber and curr_diff >= frac * part_measures_secondsMap[prev_note.measureNumber-1]['durationSeconds']):
                            prev_note.groups.append('}')
                            curr_note.groups.append('{')
                    prev_diff = curr_diff
                else:  # first note of work
                    curr_note.groups.append('{')
                prev_secondsDict = secondsDict
            # always mark last note in part
            if prev_note:
                prev_note.groups.append('}')
            # add grouping lyrics for debug
            for note in part.recurse().getElementsByClass(m21.note.GeneralNote):
                note.lyric = ' '.join(note.groups)


    def findVoltas(self):
        self.RepeatedMeasureByVolta = [0] * self.NumMeasures
        RepeatBrackets = self.NoteStream.recurse().getElementsByClass('RepeatBracket')
        for repeat in RepeatBrackets:
            for sp_store in repeat.spannerStorage:
                measureIndex = sp_store.measureNumber - 1
                self.RepeatedMeasureByVolta[measureIndex] = 1

    def findFinalBarlines(self):
        self.FinalBarlines = [0] * self.NumMeasures
        final_barlines = self.NoteStream.recurse().getElementsByClass('Barline')
        for curr_barline in final_barlines:
            if curr_barline.type == 'final':
                measureIndex = curr_barline.measureNumber - 1
                self.FinalBarlines[measureIndex] = 1

    def measureContainsFinalBarline(self, measure):
        retVal = False
        barlines = measure.recurse().getElementsByClass('Barline')
        for curr_barline in barlines:
            if curr_barline.type == 'final':
                retVal = True
                break
        return retVal

    def measureContainsVolta(self, measure):
        retVal = False
        RepeatBrackets = measure.recurse().getElementsByClass('RepeatBracket')
        for repeat in RepeatBrackets:
            for _ in repeat.spannerStorage:
                retVal = True
                break
        return retVal

    def findIncompleteMeasures(self):
        print('Finding incomplete measures...')
        self.IncompleteMeasures = [0] * self.NumMeasures
        time_sig = []
        measure_zero = self.ChordStream.measure(0)
        if measure_zero:
            for currSig in measure_zero.recurse().getElementsByClass(m21.meter.TimeSignature):
                time_sig = currSig
        for i in range(self.NumMeasures): #enumerate(self.ChordStream.recurse().getElementsByClass(stream.Measure)):
            measure = self.ChordStream.measure(i+1)
            if measure:
                if i==36:
                    bla = 0
                for timeSig in measure.recurse().getElementsByClass(m21.meter.TimeSignature):
                    time_sig = timeSig
                chords = [chord for chord in measure.recurse().getElementsByClass('GeneralNote')]
                if chords and not chords[-1].isRest: # this attemps to avoid false pickups (i.e. ignore cases where the rest is just not long enough) - not sure this is solid for all pickups
                    lengths = [chord.duration.quarterLength for chord in chords]
                    total_duration = sum(lengths)
                    epsilon = 0.1 # this epsilon is needed because of inaccuracy caused by decoration notes in some scores
                    if total_duration < (1 - epsilon) * time_sig.barDuration.quarterLength:
                        self.IncompleteMeasures[i] = True
        # using pickup detection from Haydn, find first non-empty measure and set it incomplete there exists a pickup
        first_non_empty_measure = self.EmptyMeasures.index(0)
        self.IncompleteMeasures[first_non_empty_measure] = self.IncompleteMeasures[first_non_empty_measure] or self.hasPickupMeasure
        incomplete_measures = [i+1 for i,inc in enumerate(self.IncompleteMeasures) if inc]
        print('Incomplete meaures:', incomplete_measures)


    def findEmptyMeasures(self):
        print('Finding empty measures...')
        self.EmptyMeasures = [0] * self.NumMeasures
        for i in range(self.NumMeasures):
            curr_measure = self.ChordStream.measure(i+1)
            if curr_measure and len(curr_measure.recurse().notesAndRests) == 0:
                self.EmptyMeasures[i] = 1
        empty_measures = [i + 1 for i, emp in enumerate(self.EmptyMeasures) if emp]
        print('Empty measures:', empty_measures)


    def removePickupMeasure(self):
        Parts = self.NoteStream.getElementsByClass(stream.Part)
        self.hasPickupMeasure = 1
        for part in Parts:
            if not part.flat.notesAndRests[0].isRest:
                self.hasPickupMeasure = 0
                print("Pickup Cancelled")
                break

    def addMyLablesToParts(self, note_stream):
        parts = note_stream.recurse().getElementsByClass(m21.stream.Part)
        # for more than one voice, assuming soprano is first and bass last, remaining parts are not accessed individually
        parts[0].id = 'MySoprano'
        parts[-1].id = 'MyBasso'

    def removeGraceNotes(self, note_stream):
        # remove grace notes
        note_stream = copy.deepcopy(note_stream)
        graceNotes = []
        for n in note_stream.recurse().notes:
            if n.duration.isGrace:
                graceNotes.append(n)
        for grace in graceNotes:
            grace.activeSite.remove(grace)
        return note_stream

    def tryIsChord(x, self):
        try:
            return x.isChord
        except:
            return False

    def removeChordSymbolsInNoteStream(self):
        objects_to_remove = []
        for curr_note in self.NoteStream.recurse().getElementsByClass('ChordSymbol'):
            if curr_note.isChord:
                objects_to_remove.append(curr_note)

        self.NoteStream.remove(objects_to_remove,recurse=True)

        chords_present = False
        for curr_note in self.NoteStream.recurse().getElementsByClass(note.GeneralNote):
            if curr_note.isChord:
                chords_present = True
                break
        if not chords_present:
             print('No more chords in note stream!')

    def replaceBassRestsWithPrevsViaSecondsMaps(self):
        max_consec_empty_meas = 1
        prev_note = []
        prev_measure_dur = []
        consecutive_empty_measure_durations = []
        measures_to_tie = []
        parts = self.NoteStreamRestless.recurse().getElementsByClass(m21.stream.Part)
        for curr_measure in parts[-1].recurse().getElementsByClass(m21.stream.Measure):
            notes_in_measure = 0
            measure_modifcations_list = []
            if curr_measure.measureNumber == 121:
                bla  = 0
            # get curr meaure duration before alterations
            curr_measure_duration = copy.deepcopy(curr_measure.duration.quarterLength)
            #find rest and create modifications
            sortedSecondsMaps = sorted(curr_measure.recurse().secondsMap, key=lambda d: d['offsetSeconds'])
            for curr_item in sortedSecondsMaps:
                #only notes, chords, and rests have durations
                if not curr_item['durationSeconds'] > 0:
                    continue
                if curr_item['element'].isRest:# found rest, remove it but not after one measure
                    if prev_note and (notes_in_measure > 0 or len(consecutive_empty_measure_durations) <= max_consec_empty_meas):
                        measure_modifcations_list.append(curr_item['element'])
                else: # note or chord
                    if not prev_note:
                        prev_note = curr_item
                    else: # not a rest and prev_note exists
                        if notes_in_measure > 0 and (curr_item['element'].offset == prev_note['element'].offset or curr_item['endTimeSeconds'] == prev_note['endTimeSeconds']):# if has same offset but lower bass update new curr item without changing duration
                            curr_lowest_pitch = curr_item['element'].sortFrequencyAscending().pitches[0] if curr_item['element'].isChord else curr_item['element'].pitch
                            prev_lowest_pitch = prev_note['element'].sortFrequencyAscending().pitches[0] if prev_note['element'].isChord else prev_note['element'].pitch
                            if curr_lowest_pitch.midi < prev_lowest_pitch.midi:
                                prev_note = curr_item
                        else: # found note or chord with new offset, extend duration till this note
                            curr_offset = curr_item['element'].offset
                            prev_offset = prev_note['element'].offset
                            extended_quarterLength = curr_offset - prev_offset
                            if notes_in_measure == 0 and prev_measure_dur: # first note in measure, complete previous measure
                                extended_quarterLength += prev_measure_dur
                            # now complete the empty measues (if exist)
                            extended_quarterLength += np.sum([dur for dur in consecutive_empty_measure_durations])
                            if extended_quarterLength > prev_note['element'].quarterLength and (notes_in_measure > 0 or len(consecutive_empty_measure_durations) <= max_consec_empty_meas):
                                prev_note['element'].duration = m21.duration.Duration(extended_quarterLength)
                                prev_note['element'].quarterLength = extended_quarterLength
                            # now check if prev_note exceeded its measure length
                            if (prev_note['element'].offset + prev_note['element'].quarterLength > curr_measure_duration):
                                measures_to_tie.append({'from': prev_note['element'].measureNumber, 'to': curr_measure.measureNumber})
                            prev_note = curr_item
                    notes_in_measure += 1
                    consecutive_empty_measure_durations = []
            #extend last note to complete measure
            if prev_note and notes_in_measure > 0:
                remainder = curr_measure_duration - prev_note['element'].offset
                if remainder > prev_note['element'].quarterLength:
                    prev_note['element'].duration = m21.duration.Duration(remainder)
                    prev_note['element'].quarterLength = remainder
            # check for final barline
            if curr_measure.rightBarline and curr_measure.rightBarline.type == 'final':
                prev_note = []
            # check if measure was empty
            if notes_in_measure == 0 and prev_note:
                consecutive_empty_measure_durations.append(curr_measure.duration.quarterLength)
            prev_measure_dur = curr_measure.duration.quarterLength
            # make erase modifications to measure
            curr_measure.remove(measure_modifcations_list, recurse=True)
        # making ties for notes extending beyond measure locally
        makeTies_local_failure = False
        for curr_measures_to_tie in measures_to_tie:
            try:
                self.NoteStreamRestless.parts[-1].measures(curr_measures_to_tie['from'], curr_measures_to_tie['to']).recurse().makeTies(inPlace=True)
            except:
                makeTies_local_failure = True
                print(f'makeTies failed on measures {curr_measures_to_tie}')
        # making ties on entire stream if it failed locally
        if makeTies_local_failure:
            try:
                self.NoteStreamRestless.parts[-1].recurse().makeTies(inPlace=True)
                print(f'makeTies overall succeeded')
            except:
                print(f'makeTies failed')
        else:
            print(f'makeTies locally succeeded')

    def noteFromRestWithPitch(self,rest,pitch):
        noteFromRest = m21.note.Note(pitch)
        noteFromRest.duration = rest.duration
        noteFromRest.offset = rest.offset
        noteFromRest.quarterLength = rest.quarterLength
        return noteFromRest

    def chordFromRest(self,rest):
        chordFromRest = m21.chord.Chord([])
        chordFromRest.duration = rest.duration
        chordFromRest.offset = rest.offset
        chordFromRest.quarterLength = rest.quarterLength
        return chordFromRest


    def getNumMeasures(self):
        return self.NumMeasures

    def detectKeyPerMeasure(self):
        print("Detecting key per block...")
        print("Num Measures:", self.NumMeasures)
        for currBlock in range(1, self.NumMeasures+1):
            # this only looks at current bar and backwards, simulating realtime
            lookahead = round((self.blockSize - 1) * self.key_detection_lookahead)
            lookback = self.blockSize - lookahead - 1
            StopMeasure = min(currBlock + lookahead, self.NumMeasures)
            StartMeasure = max(currBlock - lookback, 0)
            CurrMeasures = self.NoteStream.measures(StartMeasure, StopMeasure)
            #print(currBlock, StopMeasure)
            try:
                #analysisClasses = [
                #    Ambitus,
                #    KrumhanslSchmuckler,
                #    AardenEssen,
                #    SimpleWeights,
                #    BellmanBudge,
                #    TemperleyKostkaPayne,
                #]
                Key = CurrMeasures.analyze('TemperleyKostkaPayne')
                Key.alternateInterpretations.append(Key)
                CurrInterpretations = list(Key.alternateInterpretations)
            except:
                print("Key detection error:", sys.exc_info()[0])
                print("Maintaining previous key")
                CurrInterpretations = list(self.InterpretationsPerMeasure[-1])

            self.InterpretationsPerMeasure.append(CurrInterpretations)

    def setConstantKeyPerMeasure(self, key_string):
        const_key = m21.key.Key(key_string)
        const_key.correlationCoefficient = 1.0
        dummy_key_string = 'C' if key_string != 'C' else 'A'
        dummy_key = m21.key.Key(dummy_key_string)
        dummy_key.correlationCoefficient = 0.0
        const_key_list = [const_key, dummy_key]
        for m in range(1, self.NumMeasures + 1):
            self.InterpretationsPerMeasure.append(const_key_list)

    def reenforceKeyByFactor(self, keyString, enhancement):
        enhancedCoef = math.pow(self.CurrSmoothedInterpretations[keyString], 1/enhancement)
        self.CurrSmoothedInterpretations[keyString] = enhancedCoef

    def writeKeyPerMeasureToFile(self):
        if self.KeyDetectionMode != CDKeyDetectionModes.KSWithSmoothingCadenceSensitive:
            print("Writing keys to file...")
            keyfileName = self.fileName.replace(".", "_")
            FullPath = os.path.join(self.WritePath, keyfileName)
            KeyFile = f"{FullPath}_Key_{self.KeyDetectionMode}.txt"
            with open(KeyFile, 'wb') as f:
                pickle.dump(self.KeyPerMeasure, f)

    def readKeyPerMeasureFromFile(self):
        if self.KeyDetectionMode != CDKeyDetectionModes.KSWithSmoothingCadenceSensitive:
            print("Reading keys from file...")
            keyfileName = self.fileName.replace(".", "_")
            FullPath = os.path.join(self.WritePath, keyfileName)
            KeyFile = f"{FullPath}_Key_{self.KeyDetectionMode}.txt"
            with open(KeyFile, 'rb') as f:
                self.KeyPerMeasure=pickle.load(f)

    def getKeyPerMeasureFromSearsFile(self,FullPath):
        FullPath = FullPath.replace(".xml", ".txt")

        class SearKeyDataIndices(enum.IntEnum):
            Key = 0
            Mode = 1
            StartBar = 2
            StartPulse = 3
            EndBar = 4
            EndPulse = 5

        with open(FullPath,'r') as f:
            lines = f.readlines()
            FoundRow = 0
            totalNumMeasures = 0
            startMeasure = 0
            endMeasure = -1
            for line in lines:
                if FoundRow==0:
                    if "Tonal Region" in line:
                        FoundRow = 1
                        continue
                else:
                    if "Major" in line or "Minor" in line:
                        elements = line.strip().split("\t")
                        currKey = self.mapSearsKeyToMusic21Key(elements[SearKeyDataIndices.Key],elements[SearKeyDataIndices.Mode])
                        print(elements, len(elements))
                        #print(line, file=text_file_reduced)
                        startMeasure = elements[SearKeyDataIndices.StartBar]

                        if "Begin" in startMeasure:
                            startMeasure = 1
                        else:
                            startMeasure = int(startMeasure)

                        # check if start measure is before prev end measure then there is an overlap in tonal regions (TBD - what to do with this?)
                        if startMeasure<=endMeasure:
                            startMeasure = endMeasure + 1

                        endMeasure = elements[SearKeyDataIndices.EndBar]
                        if "End" in endMeasure:
                            endMeasure = self.NumMeasures
                        else:
                            endMeasure = int(endMeasure)

                        numMeasuresToAppend = endMeasure-startMeasure+1 #including start and end

                        for i in range(numMeasuresToAppend):
                            self.KeyPerMeasure.append(currKey)

            print("KeysLen=", len(self.KeyPerMeasure))
            print("NumMeasures=",self.NumMeasures)

    def mapSearsKeyToMusic21Key(self,SearStringKey, SearStringMode):
        Music21StringKey = SearStringKey.replace("b","-")
        Music21StringMode = SearStringMode.replace("M","m")
        return m21.key.Key(Music21StringKey,Music21StringMode)

    def isArpeggioPattern(self, pitches, lengths, arp_len, up_down=None):
        if len(pitches) < arp_len or not all(lengths[0] == l for l in lengths[1:-1]): # the final note in an arpeggio can be longer
            retVal = False
        else:
            retVal = True
            if up_down == 'up':
                if all(x.midi > pitches[0].midi for x in pitches[1:]):
                    prev_pitch = []
                    for p in pitches:
                        if prev_pitch and (p.midi - prev_pitch.midi) < 3: #minor and major second disqualifies the arpeggio
                            retVal = False
                            break
                        prev_pitch = p
                else:
                    retVal = False
            elif up_down == 'down':
                if all(x.midi < pitches[0].midi for x in pitches[1:]):
                    prev_pitch = []
                    for p in pitches:
                        if prev_pitch and (prev_pitch.midi - p.midi) < 3: #minor and major second disqualifies the arpeggio
                            retVal = False
                            break
                        prev_pitch = p
                else:
                    retVal = False
        return retVal

    def isAlbertiPattern(self, pitches, lengths, pattern_len):
        equal_durations = all(lengths[0] == l for l in lengths[1:4])
        lowest_pitch_first = all(x.midi >= pitches[0].midi for x in pitches[1:4])
        lowest_pitch_third = all(x.midi >= pitches[2].midi for x in pitches[1:4])
        repetitve_pitch = (pitches[1].midi == pitches[3].midi or pitches[0].midi == pitches[2].midi)
        consecutive_pitch = np.any(np.diff([x.midi for x in pitches]) == 0)
        # for pattern length 6 (3/4,3/8,6/8) verify last two as well, either continuing the same pattern or repeating the bass
        if pattern_len == 6:
            equal_durations = all(lengths[0] == l for l in lengths)
            repetitve_pitch = repetitve_pitch and (pitches[3].midi == pitches[5].midi or pitches[0] == pitches[4])
        is_alberti = equal_durations and repetitve_pitch and (lowest_pitch_first or lowest_pitch_third) and not consecutive_pitch
        bass_pitch_index = None
        if lowest_pitch_first:
            bass_pitch_index = 0
        elif lowest_pitch_third:
            bass_pitch_index = 2
        return is_alberti, bass_pitch_index

    def extractBassLine(self, measure):
        lowestPitches = []
        quarterLengths= []
        for chord in measure:
            if not chord.isRest:
                lowestNote = chord.sortFrequencyAscending()[0]
                lowestPitches.append(lowestNote.pitch)
                quarterLengths.append(lowestNote.quarterLength)
            else:
                lowestPitches.append(None)
        return {'pitches': lowestPitches, 'lengths': quarterLengths}

    def detectAlbertiOrArpeggioBassInMeasure(self, measure_chords, bass_chords, pattern_len, arp_type):
        #first exact bassline from chords
        measure_general_notes = measure_chords.recurse().getElementsByClass('GeneralNote')
        bass_general_notes = bass_chords.recurse().getElementsByClass('GeneralNote')
        lowest_pitches_in_bass = self.extractBassLine(bass_general_notes)
        arp_patterns = []

        # need to loop like this because we advance the index based on finding arpeggios
        i = 0
        while i <= len(bass_general_notes)-pattern_len:
            curr_pitches = lowest_pitches_in_bass['pitches'][i:i+pattern_len]
            curr_lengths = lowest_pitches_in_bass['lengths'][i:i+pattern_len]
            isArp = False
            if None not in curr_pitches:
                if any([cl > 1 for cl in curr_lengths]): # do not consider arps greater than quarter notes
                    isArp = False
                elif arp_type == 'alberti':
                    isArp, bass_pitch_index = self.isAlbertiPattern(pitches=curr_pitches, lengths=curr_lengths, pattern_len=pattern_len)
                elif arp_type == 'arp_up':
                    isArp = self.isArpeggioPattern(pitches=curr_pitches, lengths=curr_lengths, arp_len=pattern_len, up_down='up')
                    bass_pitch_index = 0
                elif arp_type =='arp_down':
                    isArp = self.isArpeggioPattern(pitches=curr_pitches, lengths=curr_lengths, arp_len=pattern_len, up_down='down')
                    bass_pitch_index = 0
                else:
                    isArp = False
                if isArp:
                    epsilon = 0.1  # this epsilon is required to mark the last alberti beat including its length, but not the next beat
                    # already assuming they are all equal length
                    beat_diff = bass_general_notes[i+1].beat-bass_general_notes[i].beat # for some reason quarter length does not always provide this, TBD use seconds map
                    if bass_pitch_index == 0:
                        arp_patterns.append({'start': bass_general_notes[i + 1].beat,
                                             'stop': bass_general_notes[i + pattern_len - 1].beat + beat_diff - epsilon})
                    else:
                        # mark from beginning of pattern not including first note, to bass pitch index not included
                        arp_patterns.append({'start': bass_general_notes[i + 1].beat,
                                             'stop': bass_general_notes[i + bass_pitch_index - 1].beat + beat_diff - epsilon})
                        # mark from note after bass pitch index to the end
                        arp_patterns.append({'start': bass_general_notes[i + bass_pitch_index + 1].beat,
                                             'stop': bass_general_notes[i + pattern_len - 1].beat + beat_diff - epsilon})
            # advance index by pattern length if alberti found, otherwise advance to next beat
            i = i + pattern_len if isArp else i + 1


        arp_beats = [0] * len(measure_general_notes)
        for i,gen_note in enumerate(measure_general_notes):
            for pat in arp_patterns:
                if pat['start'] <= gen_note.beat <= pat['stop']:
                    arp_beats[i] = True
                    break

        return arp_beats

    def smoothKeyInterpretations(self, measureIndex):
        currInterpretations = self.InterpretationsPerMeasure[measureIndex]
        for interpretation in currInterpretations:
            corrCoef = interpretation.correlationCoefficient
            keyString = interpretation.tonicPitchNameWithCase
            if keyString not in self.CurrSmoothedInterpretations:
                self.CurrSmoothedInterpretations[keyString] = corrCoef
            currSmoothedVal = self.CurrSmoothedInterpretations[keyString]
            currSmoothedVal = self.KeyDetectionForgetFactor * currSmoothedVal + (1-self.KeyDetectionForgetFactor) * corrCoef
            self.CurrSmoothedInterpretations[keyString] = currSmoothedVal

    def resetSmoothedInterpretations(self):
        self.CurrSmoothedInterpretations = {}

    def getTopNKeyInterpretations(self, N):
        #now sort by value
        SortedKeys = sorted(self.CurrSmoothedInterpretations.items(), key=lambda x: x[1], reverse=True)
        TopNKeys = list(SortedKeys[:N])
        return TopNKeys


    def getNotesPerChordBeatPerPart(self, ChordsMeasure, NotesMeasure):
        Parts = NotesMeasure.recurse().getElementsByClass(m21.stream.Part)
        parts_flat = [part.flat.notesAndRests for part in Parts]
        notesPerChordBeatPerPart = [[[real_note for real_note in part if chord.beat == real_note.beat] for part in parts_flat] for chord in ChordsMeasure.recurse().getElementsByClass('GeneralNote')]
        return notesPerChordBeatPerPart


    def detectCadences(self):
        print("Detecting cadences...")
        self.analysis_full = []
        self.analysis_transition_strings = []

        measuresSecondsMap = list(filter(lambda d: d['durationSeconds'] > 0, self.ChordStreamRestless.secondsMap))

        prevState = []
        repeat_counter = 0
        incomplete_counter = 0
        empty_measure_counter = 0
        coefs_per_measure = []

        # check first timesig
        measure_zero = self.ChordStream.measure(0)
        curr_time_sig = self.check_and_update_timesig(measure_zero, [])

        # loop on measures:
        try:
            for currMeasureIndex in range(0,self.NumMeasures):
                # debug per measure
                if currMeasureIndex == 18:
                    bla = 0
                # true measures start with 1, pickups will start from zero, but not all corpora will abide to this
                # for example, data that originates from midi cannot contain this info
                # to overcome this, we attempt to find the pickup via initial rests and discard it
                # also, we count the empty measures and index them out while writing the label
                measure_number = currMeasureIndex + 1
                # print(f"measure number {measure_number}")
                CurrMeasuresRestless = self.ChordStreamRestless.measure(measure_number)
                CurrMeasures = self.ChordStream.measure(measure_number)
                CurrMeasuresNotes = self.NoteStream.measure(measure_number)
                CurrMeasureBass = self.BassChords.measure(measure_number)

                # check and update timesig
                curr_time_sig = self.check_and_update_timesig(CurrMeasures, curr_time_sig)

                # reset state machines after repeat brackets, unless cadence inevitable (TBD  - add double barlines to this)
                if currMeasureIndex > 0 and self.FinalBarlines[currMeasureIndex-1] and not (self.HarmonicStateMachine.getCadentialOutput() == CDCadentialStates.CadInevitable):
                    self.HarmonicStateMachine.reset()
                    self.HarmonicStateMachineChallenger.reset()
                    self.resetSmoothedInterpretations()

                if self.EmptyMeasures[currMeasureIndex]:
                    empty_measure_counter = empty_measure_counter + 1
                    continue

                if self.IncompleteMeasures[currMeasureIndex]:
                    incomplete_counter = incomplete_counter + 1

                if not CurrMeasures or not CurrMeasuresRestless:
                    continue

                NotesPerChordBeatPerPart = self.getNotesPerChordBeatPerPart(CurrMeasures, CurrMeasuresNotes)

                AlbertiBeats4 = self.detectAlbertiOrArpeggioBassInMeasure(CurrMeasuresRestless, CurrMeasureBass, pattern_len=4, arp_type='alberti')
                AlbertiBeats = AlbertiBeats4
                if curr_time_sig.ratioString in ['3/4', '6/8', '3/8']:
                    AlbertiBeats6 = self.detectAlbertiOrArpeggioBassInMeasure(CurrMeasuresRestless, CurrMeasureBass, pattern_len=6, arp_type='alberti')
                    AlbertiBeats = [x or y for (x, y) in zip(AlbertiBeats4, AlbertiBeats6)]

                # consider Alberti only after the second consecutive measure
                # AlbertiBeatsFiltered = [x and self.PrevMeasureAlberti for x in AlbertiBeats]
                # self.PrevMeasureAlberti = any(AlbertiBeats)

                ArpeggioBeats4 = self.detectAlbertiOrArpeggioBassInMeasure(CurrMeasuresRestless, CurrMeasureBass, pattern_len=4, arp_type='arp_up')
                ArpeggioBeats = ArpeggioBeats4
                if curr_time_sig.ratioString in ['3/4', '6/8', '3/8']:
                    ArpeggioBeats3 = self.detectAlbertiOrArpeggioBassInMeasure(CurrMeasuresRestless, CurrMeasureBass, pattern_len=3, arp_type='arp_up')
                    ArpeggioBeats = [x or y for (x, y) in zip(ArpeggioBeats3, ArpeggioBeats4)]

                # consider Arpeggio only after the second consecutive measure
                ArpeggioBeatsFiltered = [x and self.PrevMeasureArpeggio for x in ArpeggioBeats]
                self.PrevMeasureArpeggio = any(ArpeggioBeats)

                LyricPerBeat = []

                if self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothing or\
                        self.KeyDetectionMode==CDKeyDetectionModes.KSWithSmoothingCadenceSensitive:
                    #smooth keys here, per measure (TBD - per chord?):
                    self.smoothKeyInterpretations(currMeasureIndex)
                    #sort and return top 2
                    self.Top2Keys = self.getTopNKeyInterpretations(2)
                else:
                    currKey = self.KeyPerMeasure[currMeasureIndex]#lists start with 0

                for thisChord, thisChordWithBassRests, alberti, arpeggio, thisRealNotes in zip(CurrMeasuresRestless.recurse().getElementsByClass('GeneralNote'),CurrMeasures.recurse().getElementsByClass('GeneralNote'), AlbertiBeats, ArpeggioBeatsFiltered, NotesPerChordBeatPerPart):
                    if self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothing or\
                            self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothingCadenceSensitive:
                        currKeyString = self.Top2Keys[0][0]
                        currKey = m21.key.Key(currKeyString)
                        challengerKeyString = self.Top2Keys[1][0]
                        challengerKey = m21.key.Key(challengerKeyString)
                        self.updateKeysAndSwapStateMachines(currKeyString, challengerKeyString)

                    # main key state machine
                    if thisChordWithBassRests.isRest:
                        self.HarmonicStateMachine.updateHarmonicState(currKey, thisChord, thisChordWithBassRests, [], [], [], alberti, arpeggio, [], thisRealNotes)
                    else:
                        rn = m21.roman.romanNumeralFromChord(thisChordWithBassRests, currKey)
                        self.HarmonicStateMachine.updateHarmonicState(currKey, thisChord, thisChordWithBassRests, rn.scaleDegree, rn.inversion(), rn.figure, alberti, arpeggio, rn, thisRealNotes)

                    # challenger key state machine
                    if self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothing or self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothingCadenceSensitive:
                        if thisChordWithBassRests.isRest:
                            self.HarmonicStateMachineChallenger.updateHarmonicState(challengerKey, thisChord, thisChordWithBassRests, [], [], [], alberti, arpeggio, [], thisRealNotes)
                        else:
                            rn2 = m21.roman.romanNumeralFromChord(thisChordWithBassRests, challengerKey)
                            self.HarmonicStateMachineChallenger.updateHarmonicState(challengerKey, thisChord, thisChordWithBassRests, rn2.scaleDegree, rn2.inversion(), rn2.figure, alberti, arpeggio, rn2, thisRealNotes)

                    #TBD - should we handle this reversion in challenger?
                    if self.HarmonicStateMachine.getRevertLastPACAndReset() and len(LyricPerBeat) > 0: #this limits PAC reversion to within measure
                        LastPACTuple = LyricPerBeat[-1]
                        UpdatedLyric = "IAC"
                        LyricPerBeat[-1] = [LastPACTuple[0], UpdatedLyric]
                        print("Reverting last PAC to IAC")
                    if self.HarmonicStateMachine.getRevertLastHCAndReset() and len(LyricPerBeat) > 0:  # this limits HC reversion to within measure
                        LastHCTuple = LyricPerBeat[-1]
                        UpdatedLyric = ""
                        LyricPerBeat[-1] = [LastHCTuple[0], UpdatedLyric]
                        print("Reverting last HC")
                    if self.HarmonicStateMachine.getRevertLastIACAndReset() and len(LyricPerBeat) > 0:  # this limits IAC reversion to within measure
                        LastIACTuple = LyricPerBeat[-1]
                        UpdatedLyric = ""
                        LyricPerBeat[-1] = [LastIACTuple[0], UpdatedLyric]
                        print("Reverting last IAC")

                    Lyric = self.HarmonicStateMachine.getCadentialOutputString()
                    LyricChallenger = self.HarmonicStateMachineChallenger.getCadentialOutputString()
                    # Cadence Sensitive Key Detection Mode
                    if self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothingCadenceSensitive:
                        # and self.HarmonicStateMachineChallenger.CurrHarmonicState.Key.tonic != self.HarmonicStateMachine.CurrHarmonicState.Key.tonic: #don't be sensitive to candece in parallel key:
                        # Cadence check in main key (for key re-enforcement)

                        resort = False
                        for cad in self.ReenforcementFactors:
                            if cad in Lyric:
                                # avoid reinforcement of HC if challenger is on dominant
                                reenforce = not (currKey.getDominant().pitchClass == challengerKey.tonic.pitchClass and challengerKey.mode=='major' and (cad == 'HC'))
                                if reenforce:
                                    self.reenforceKeyByFactor(currKeyString, self.ReenforcementFactors[cad])
                            # Challenger Check (for key switching, but ignore HCs in subdominant key (becuase they are more likely PACs in main key))
                            if cad in LyricChallenger:
                                reenforce =  not (challengerKey.tonic.pitchClass == self.getSubDominantPitchClass(currKey) and challengerKey.mode=='major' and (cad == 'HC'))
                                if reenforce:
                                    self.reenforceKeyByFactor(challengerKeyString, self.ReenforcementFactors[cad])
                                    resort = True
                        if resort:
                            # re-sort and return Top2Keys
                            self.Top2Keys = self.getTopNKeyInterpretations(2)
                            if challengerKeyString == self.Top2Keys[0][0]:
                                Lyric = LyricChallenger
                                print('Cadential Key Change!')

                    # debugging
                    # print(self.HarmonicStateMachine.getCadentialOutput().value)
                    # thisChord.lyric = str(rn.figure)

                    if Lyric:   # only work on non-empty lyrics
                        thisChord.lyric = Lyric
                        LyricPerBeat.append([thisChord.beat,Lyric])
                        #TBD - do we still need to print this?
                        self.analysis_transition_strings.append(f"{measuresSecondsMap[currMeasureIndex]['offsetSeconds'] + measuresSecondsMap[currMeasureIndex]['durationSeconds']*(thisChord.beat-1)/CurrMeasuresRestless.duration.quarterLength:.1f} {self.HarmonicStateMachine.getCadentialOutput().value}")

                #checking cadences on entire measure
                for LyricTuple in LyricPerBeat:
                    Lyric = LyricTuple[1]
                    LyricsToPrint = ["PAC", "IAC", "HC", "PCC", "PCH"]
                    if any([s in Lyric for s in LyricsToPrint]):
                        # incomplete and empty counters count measures that should be discarded
                        measure_to_write = measure_number - self.hasPickupMeasure - empty_measure_counter
                        print('Measure ', measure_to_write, ' offset ', measuresSecondsMap[currMeasureIndex]['offsetSeconds'], " ", Lyric)
                        self.analysis_full.append({"Measure": measure_to_write, "Offset": measuresSecondsMap[currMeasureIndex]['offsetSeconds'], "Lyric": Lyric})
                    elif "Key" in Lyric:
                        print(Lyric)

                # ONLY NOW update repeat counter!!
                repeat_counter = repeat_counter + self.RepeatedMeasureByVolta[currMeasureIndex]

                    # save transitions to file regardless of Lyrics
                    # if prevState != self.HarmonicStateMachine.getCadentialOutput().value:
                    #    print(f"{measuresSecondsMap[currMeasureIndex]['offsetSeconds']} {self.HarmonicStateMachine.getCadentialOutput().value}", file=text_fileTransitions)
                    # prevState = self.HarmonicStateMachine.getCadentialOutput().value
                    # thisChord.lyric = str("block ") + str(i)

                # === for debugging measure number problems
                #LyricPerBeat.append((1,str(currMeasureIndex)))
                # === for debugging key shifting
                #currKeyString = self.Top2Keys[0][0]
                #challengerKeyString = self.Top2Keys[1][0]
                #LyricPerBeat.append((1, f'{currKeyString} {self.CurrSmoothedInterpretations[currKeyString]}'))
                #LyricPerBeat.append((1, f'{challengerKeyString} {self.CurrSmoothedInterpretations[challengerKeyString]}'))

                # write coefs per measure for offline analysis
                coefs_per_measure.append(copy.deepcopy(self.CurrSmoothedInterpretations))

                Parts = CurrMeasuresNotes.recurse().getElementsByClass(m21.stream.Part)
                Lyrics_to_filter = ['CA', '{', '}'] # for filtering the display
                for thisLyric in LyricPerBeat:
                    if thisLyric[1] not in Lyrics_to_filter:
                        found = 0
                        for index in range(0, self.NumParts):
                            for thisNote in Parts[-1-index].flat.notesAndRests:  # adding lyric from last part up, assuming its the bass line
                                if thisNote.beat == thisLyric[0]:
                                    thisNote.lyric = thisLyric[1]
                                    found = 1
                                    break
                            if found == 1:
                                break

                #write corrected key at end of measure
                if self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothingCadenceSensitive or self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothing:
                    correctedKeyString = self.Top2Keys[0][0]
                    correctedKey = m21.key.Key(correctedKeyString)
                    self.CorrectedKeyPerMeasure.append(correctedKey)
        except:
            print(f"Exception occurred on measure {measure_number}")
            raise

        for c in self.ChordStreamRestless.recurse().getElementsByClass('Chord'):
            c.closedPosition(forceOctave=4, inPlace=True)
        # this adds the restless chord stream to the note stream
        # self.NoteStream.insert(0, self.ChordStream)
        self.filter_notestream_by_language_model()
        self.filter_analysis_by_language_model()
        self.cleanup_lyrics(['{', '}'])
        self.write_analysis_to_files()

        if self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothingCadenceSensitive or self.KeyDetectionMode == CDKeyDetectionModes.KSWithSmoothing:
            print("Writing cadence corrected keys to file...")
            keyfileName = self.fileName.replace(".", "_")
            FullPath = os.path.join(self.WritePath, keyfileName)
            KeyFile = f"{FullPath}_Key_{self.KeyDetectionMode}.txt"
            with open(KeyFile, 'wb') as f:
                pickle.dump(self.CorrectedKeyPerMeasure, f)
            CoefsFile = f"{FullPath}_Coefs_{self.KeyDetectionMode}.txt"
            with open(CoefsFile, 'wb') as f:
                pickle.dump(coefs_per_measure, f)

    def isilos(self, sub, list_or_string):
        if isinstance(list_or_string, list):
            ret_val = any(sub in s for s in list_or_string)
        else:
            ret_val = sub in list_or_string
        return ret_val

    def cleanup_lyrics(self, lyric_list):
        print('cleaning up lyrics')
        for part in self.Parts:
            for measure in part.recurse().getElementsByClass(m21.stream.Measure):
                for thisNote in measure.recurse().getElementsByClass(m21.note.GeneralNote):
                    for lyric in lyric_list:
                        thisNote.lyric = thisNote.lyric.replace(lyric, '')

    def filter_notestream_by_language_model(self):
        print('Filtering lyrics by language model...')
        PrevMeasureNotes = None
        PrevMeasureLyrics = None
        AnchorCadenceMeasureNumber = 0
        PrevParts = []
        empty_measure_counter = 0
        for currMeasureIndex in range(0, self.NumMeasures):
            if self.EmptyMeasures[currMeasureIndex]:
                empty_measure_counter = empty_measure_counter + 1
                continue
            measure_number = currMeasureIndex + 1
            CurrMeasureNotes = self.NoteStream.measure(measure_number)
            CurrParts = CurrMeasureNotes.recurse().getElementsByClass(m21.stream.Part)

            measure_number_for_score = measure_number - self.hasPickupMeasure - empty_measure_counter
            even_meas_since_anchor = (measure_number_for_score - AnchorCadenceMeasureNumber) % 2 == 0

            CurrMeasureLyrics = []
            for part in CurrParts:
                for thisNote in part.flat.notesAndRests:
                    CurrMeasureLyrics.append(thisNote.lyric)

            def remove_lyric_from_parts(lyric_list, parts):
                for part in parts:
                    for thisNote in part.flat.notesAndRests:
                        for lyric in lyric_list:
                            thisNote.lyric = thisNote.lyric.replace(lyric, '')

            if PrevMeasureNotes and PrevMeasureLyrics:
                # HC --> PAC
                if (self.isilos('HC', PrevMeasureLyrics) or self.isilos('PCH', PrevMeasureLyrics)) and (self.isilos('PAC', CurrMeasureLyrics) or self.isilos('PCC', CurrMeasureLyrics)):
                    if even_meas_since_anchor:
                        print(f'Filtering HC in measure {measure_number_for_score-1}')
                        remove_lyric_from_parts(['HC', 'PCH'], PrevParts)
                # PAC --> HC
                if self.isilos('PAC', PrevMeasureLyrics) and (self.isilos('HC', CurrMeasureLyrics) or self.isilos('PCH', CurrMeasureLyrics)):
                    if not even_meas_since_anchor:
                        print(f'Filtering HC in measure {measure_number_for_score}')
                        remove_lyric_from_parts(['HC', 'PCH'], CurrParts)
                # PAC + HC
                if (self.isilos('HC', CurrMeasureLyrics) or self.isilos('PCH', CurrMeasureLyrics)) and (self.isilos('PAC', CurrMeasureLyrics) or self.isilos('PCC', CurrMeasureLyrics)):
                    print(f'Filtering HC in measure {measure_number_for_score}')
                    remove_lyric_from_parts(['HC', 'PCH'], CurrParts)
                # HC --> IAC
                if (self.isilos('HC', PrevMeasureLyrics) or self.isilos('PCH', PrevMeasureLyrics)) and (self.isilos('IAC', CurrMeasureLyrics)):
                    if even_meas_since_anchor:
                        print(f'Filtering HC in measure {measure_number_for_score - 1}')
                        remove_lyric_from_parts(['HC','PCH'], PrevParts)
                    else:
                        print(f'Filtering IAC in measure {measure_number_for_score}')
                        remove_lyric_from_parts(['IAC'], CurrParts)
                # IAC --> HC
                if self.isilos('IAC', PrevMeasureLyrics) and (self.isilos('HC', CurrMeasureLyrics) or self.isilos('PCH', CurrMeasureLyrics)):
                    if even_meas_since_anchor:
                        print(f'Filtering IAC in measure {measure_number_for_score - 1}')
                        remove_lyric_from_parts(['IAC'], PrevParts)
                    else:
                        print(f'Filtering HC in measure {measure_number_for_score}')
                        remove_lyric_from_parts(['HC', 'PCH'], CurrParts)
                # HC+IAC
                if (self.isilos('HC', CurrMeasureLyrics) or self.isilos('PCH', CurrMeasureLyrics)) and self.isilos('IAC', CurrMeasureLyrics):
                    print(f'Filtering IAC in measure {measure_number_for_score}')
                    remove_lyric_from_parts(['IAC'], CurrParts)
                # PAC+IAC
                if (self.isilos('PAC', CurrMeasureLyrics) or self.isilos('PCC', CurrMeasureLyrics)) and self.isilos('IAC', CurrMeasureLyrics):
                    print(f'Filtering IAC in measure {measure_number_for_score}')
                    remove_lyric_from_parts(['IAC'], CurrParts)
                # PAC --> IAC
                if (self.isilos('PAC', PrevMeasureLyrics) or self.isilos('PCC', PrevMeasureLyrics)) and self.isilos('IAC', CurrMeasureLyrics):
                    print(f'Filtering IAC in measure {measure_number_for_score}')
                    remove_lyric_from_parts(['IAC'], CurrParts)

            # get lyrics again after filtering
            CurrMeasureLyrics = []
            for part in CurrParts:
                for thisNote in part.flat.notesAndRests:
                    CurrMeasureLyrics.append(thisNote.lyric)
            PrevMeasureLyrics = []
            for part in PrevParts:
                for thisNote in part.flat.notesAndRests:
                    PrevMeasureLyrics.append(thisNote.lyric)

            # update anchor cadence from prev part
            anchor_cad_strings = ['PAC', 'PCC']
            if PrevMeasureLyrics and any(self.isilos(s, PrevMeasureLyrics) for s in anchor_cad_strings):
                AnchorCadenceMeasureNumber = measure_number_for_score - 1

            PrevMeasureNotes = CurrMeasureNotes
            PrevMeasureLyrics = CurrMeasureLyrics
            PrevParts = CurrParts


    def filter_analysis_by_language_model(self):
        print('Filtering analysis results by language model...')
        self.analysis_full_filtered = copy.deepcopy(self.analysis_full)
        items_to_remove = []
        prev_cad = {'Measure':0, 'Lyric':[]}
        anchor_cad = {'Measure':0, 'Lyric':[]}
        for item in self.analysis_full_filtered:
            curr_cad = item
            # check if cadence in analysis line
            cad_strings = ['PAC', 'PCC', 'HC', 'PCH', 'IAC']
            if not any(s in curr_cad['Lyric'] for s in cad_strings):
                continue
            even_meas_since_anchor = (curr_cad['Measure'] - anchor_cad['Measure']) % 2 == 0
            if curr_cad['Measure'] - prev_cad['Measure'] <= 1:
                # HC --> PAC and HC+PAC
                if ('HC' in prev_cad['Lyric'] or 'PCH' in prev_cad['Lyric']) and ('PAC' in curr_cad['Lyric'] or 'PCC' in curr_cad['Lyric']):
                    if even_meas_since_anchor:
                        if prev_cad not in items_to_remove:
                            print(f"Filtering HC in measure {prev_cad['Measure']}")
                            items_to_remove.append(prev_cad)
                # PAC --> HC and PAC+HC
                if ('PAC' in prev_cad['Lyric'] or 'PCC' in prev_cad['Lyric']) and ('HC' in curr_cad['Lyric'] or 'PCH' in curr_cad['Lyric']):
                    if not even_meas_since_anchor or curr_cad['Measure'] == prev_cad['Measure']:
                        if curr_cad not in items_to_remove:
                            print(f"Filtering HC in measure {curr_cad['Measure']}")
                            items_to_remove.append(curr_cad)
                # HC --> IAC
                if ('HC' in prev_cad['Lyric'] or 'PCH' in prev_cad['Lyric']) and 'IAC' in curr_cad['Lyric']:
                    if even_meas_since_anchor and curr_cad['Measure'] > prev_cad['Measure']:
                        if prev_cad not in items_to_remove:
                            print(f"Filtering HC in measure {prev_cad['Measure']}")
                            items_to_remove.append(prev_cad)
                    else:
                        if curr_cad not in items_to_remove:
                            print(f"Filtering IAC in measure {curr_cad['Measure']}")
                            items_to_remove.append(curr_cad)
                # IAC --> HC
                if 'IAC' in prev_cad['Lyric'] and ('HC' in curr_cad['Lyric'] or 'PCH' in curr_cad['Lyric']):
                    if even_meas_since_anchor or curr_cad['Measure'] == prev_cad['Measure']:
                        if prev_cad not in items_to_remove:
                            print(f"Filtering IAC in measure {prev_cad['Measure']}")
                            items_to_remove.append(prev_cad)
                    else:
                        if curr_cad not in items_to_remove:
                            print(f"Filtering HC in measure {curr_cad['Measure']}")
                            items_to_remove.append(curr_cad)
                # IAC --> PAC
                if 'IAC' in prev_cad['Lyric'] and ('PAC' in curr_cad['Lyric'] or 'PCC' in curr_cad['Lyric']):
                    if prev_cad not in items_to_remove:
                        print(f"Filtering IAC in measure {prev_cad['Measure']}")
                        items_to_remove.append(prev_cad)
                # PAC --> IAC
                if 'IAC' in curr_cad['Lyric'] and ('PAC' in prev_cad['Lyric'] or 'PCC' in prev_cad['Lyric']):
                    if curr_cad not in items_to_remove:
                        print(f"Filtering IAC in measure {curr_cad['Measure']}")
                        items_to_remove.append(curr_cad)


            # update cadence only if not removed
            if curr_cad not in items_to_remove:
                if prev_cad not in items_to_remove:
                    # update anchor cadence only on PAC and HC
                    anchor_cad_strings = ['PAC', 'PCC']
                    if any(s in prev_cad['Lyric'] for s in anchor_cad_strings):
                        anchor_cad = prev_cad
                # update prev cad always
                prev_cad = curr_cad

        for item in items_to_remove:
            self.analysis_full_filtered = list(filter(lambda a: a != item, self.analysis_full_filtered))

    def write_analysis_to_files(self):
        fileName = self.fileName.replace(".", "_")
        FullPath = os.path.join(self.WritePath, fileName)
        text_file = open(f"{FullPath}_Analyzed.txt", "w")
        print(f"NumMeasures: {self.NumMeasures}", file=text_file)
        text_fileOffsets = open(f"{FullPath}_Analyzed_OffsetsNums.txt", "w")
        text_fileTransitions = open(f"{FullPath}_Analyzed_Transitions.txt", "w")

        for item in self.analysis_full_filtered:
            print(f"Measure: {item['Measure']} Offset: {item['Offset']} {item['Lyric']}", file=text_file)
            Lyric = item['Lyric']
            if "PAC" in Lyric:
                CadString = "PAC"
            elif "IAC" in Lyric:
                CadString = "IAC"
            elif "HC" in Lyric:
                CadString = "HC"
            elif "PCC" in Lyric:
                CadString = "PCC"
            elif "PCH" in Lyric:
                CadString = "PCH"
            CadStringToNumMap = {"PAC": CDCadentialStates.PACArrival.value, "IAC": CDCadentialStates.IACArrival.value,
                                 "HC": CDCadentialStates.HCArrival.value, "PCC": CDCadentialStates.PCCArrival.value,
                                 "PCH": CDCadentialStates.PCHArrival.value}
            print(f"{item['Measure']} {item['Offset']} {CadStringToNumMap[CadString]}", file=text_fileOffsets)

        for item in self.analysis_transition_strings:
            print(item, file=text_fileTransitions)

        text_file.close()
        text_fileOffsets.close()
        text_fileTransitions.close()

    def check_and_update_timesig(self, CurrMeasures, measure_time_sig):
        if CurrMeasures:
            for timeSig in CurrMeasures.recurse().getElementsByClass(m21.meter.TimeSignature):
                measure_time_sig = timeSig
                if timeSig != self.HarmonicStateMachine.CurrHarmonicState.TimeSig:
                    self.HarmonicStateMachine.CurrHarmonicState.TimeSig = timeSig
                    self.HarmonicStateMachineChallenger.CurrHarmonicState.TimeSig = timeSig
        return measure_time_sig

    def getSubDominantPitchClass(self, currKey):
        return (currKey.getDominant().pitchClass - 2) % 12

    def updateKeysAndSwapStateMachines(self, currKeyString, challengerKeyString):
        # this swap assures the smoothness of key transition if challenger becomes main (and main becomes challenger)
        if currKeyString != self.PrevKeyString:
            tempStateMachine = copy.deepcopy(self.HarmonicStateMachine)
            if currKeyString == self.PrevChallengerKeyString:
                self.HarmonicStateMachine = copy.deepcopy(self.HarmonicStateMachineChallenger)
                self.HarmonicStateMachine.CadentialKeyChange = 1
            if challengerKeyString == self.PrevKeyString:
                self.HarmonicStateMachineChallenger = copy.deepcopy(tempStateMachine)
            # reset post cadential counters
            self.HarmonicStateMachine.resetPostCadentialCounters()
            self.HarmonicStateMachineChallenger.resetPostCadentialCounters()
        self.PrevKeyString = currKeyString
        self.PrevChallengerKeyString = challengerKeyString
        self.HarmonicStateMachine.IsChallenger = False
        self.HarmonicStateMachineChallenger.IsChallenger = True

    def setFileName(self,fileName):
        self.fileName = fileName

    def display(self):
        # open in MuseScore
        self.ChordStream.show()

    def writeAnalyzedFile(self):
        max_num_measures_for_writing = 400
        if self.NumMeasures <= max_num_measures_for_writing:
            fileName = self.fileName.replace(".", "_")
            fileName = (f"{fileName}_Analyzed.xml")
            FullPath = os.path.join(self.WritePath, fileName)
            self.NoteStream.write(fp=FullPath)
        else:
            print(f'file too long for writing. Max num measures for writing is {max_num_measures_for_writing}.'
                  f'Num measures is:', self.NumMeasures)

    def displayFull(self):
        self.NoteStream.show()

    def setWritePath(self,path):
        self.WritePath = path



from .CadenceDetectData import *
from music21 import *
import numpy as np
import copy

class CDStateMachine(object):
    def __init__(self, isChallenger=False, minInitialMeasures=0, minPostCadenceMeasures=0):
        self.NumParts = 0
        self.MeasureCounter = 0 # measure counter is for the entire movement and not included in reset
        self.MinInitialMeasures = {CDCadentialStates.PACArrival: minInitialMeasures, CDCadentialStates.IACArrival: minInitialMeasures, CDCadentialStates.HCArrival: round(minInitialMeasures/2)}
        self.MinPostCadenceMeasures = minPostCadenceMeasures
        self.MinPostCadenceMeasuresHC = minPostCadenceMeasures+1
        self.IsChallenger = isChallenger
        self.reset()

    def reset(self):
        # initiliaze with invalid states
        self.PrevCadentialState = CDCadentialStates.Idle
        self.CurrCadentialState = CDCadentialStates.Idle
        self.CurrCadentialOutput = CDCadentialStates.Idle
        self.CurrHarmonicState = CDHarmonicState([], chord.Chord(), chord.Chord(), 0, 0, 0, 0, 0, meter.TimeSignature(), [], [])
        self.PrevHarmonicState = self.CurrHarmonicState
        self.LastNonAlbertiHarmonicState = self.CurrHarmonicState
        self.ChangeFlagOneShot = 0
        self.KeyChangeOneShot = 0
        self.FirstKeyDetectionDone = 0
        self.CadentialKeyChange = 0
        self.TriggerString = str("")
        self.PostPACMeasureCounter = self.MinPostCadenceMeasures
        self.PostHCMeasureCounter = self.MinPostCadenceMeasuresHC
        self.CheckBassPartFromChord = False
        self.PACPending = False
        self.RevertLastPAC = False
        self.HCPending = False
        self.RevertLastHC = False
        self.RevertLastIAC = False
        self.HarmonicStateOfLastCadence = 0
        self.SopranoOfLastCadence = 0
        self.WeightOfLastCadence = 0
        self.PrevMeasureAlberti = False
        self.PrevMeasureArpeggio = False

    def resetPostCadentialCounters(self):
        self.PostPACMeasureCounter = self.MinPostCadenceMeasures
        self.PostHCMeasureCounter = self.MinPostCadenceMeasuresHC

    def updateHarmonicState(self, Key, Chord, ChordWithRests, ChordDegree, ChordInversion, ChordFigure, Alberti, Arpeggio, RomamNumeral, RealNotes):

        if not (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio):
            self.LastNonAlbertiHarmonicState = copy.deepcopy(self.CurrHarmonicState)
        self.PrevHarmonicState = copy.deepcopy(self.CurrHarmonicState)
        self.KeyChangeOneShot = 0
        if self.CurrHarmonicState.Key and self.CurrHarmonicState.Key != Key:
            self.KeyChangeOneShot = 1
        self.CurrHarmonicState.Key = Key
        self.CurrHarmonicState.Chord = Chord
        self.CurrHarmonicState.ChordWithBassRests = ChordWithRests
        self.CurrHarmonicState.ChordDegree = ChordDegree
        self.CurrHarmonicState.ChordInversion = ChordInversion
        self.CurrHarmonicState.ChordFigure = ChordFigure
        self.CurrHarmonicState.Alberti = Alberti
        self.CurrHarmonicState.Arpeggio = Arpeggio
        self.CurrHarmonicState.RomanNumeral = RomamNumeral
        self.CurrHarmonicState.RealNotes = RealNotes
        self.updateCounters()
        self.updateCadentialState()

    def checkStateChanged(self):
        self.ChangeFlagOneShot = 0
        if self.PrevCadentialState != self.CurrCadentialState:
            self.ChangeFlagOneShot = 1
            self.TriggerString = str("T") + str(self.PrevCadentialState.value) + str(u'\u208B') + str(self.CurrCadentialState.value)
            self.TriggerString = self.TriggerString.translate(SUB)
            self.PrevCadentialState = self.CurrCadentialState
            if self.CurrCadentialState == CDCadentialStates.PACArrival or self.CurrCadentialState == CDCadentialStates.PCCArrival:
                self.PostPACMeasureCounter = -1
            elif self.CurrCadentialState == CDCadentialStates.HCArrival or self.CurrCadentialState == CDCadentialStates.PCHArrival:
                self.PostHCMeasureCounter = -1

    def getBassFromChord(self, chord, favor_by_part=False):
        retVal = False
        if not chord.isRest:
            # first checking the bass part (even if it is not the lowest note in the chord)
            for p in chord:
                if 'MyBasso' in p.groups:
                    if not retVal or p.pitch.midi < retVal.midi:
                        retVal = p.pitch
            if not favor_by_part or not retVal:
                # now checking the lowest pitch while verifying an octave below soprano
                sorted_pitches = chord.sortFrequencyAscending().pitches
                lowestPitch = sorted_pitches[0]
                # if MyBasso not found or if it was found but it isn't the lowest pitch
                if not retVal or retVal.midi > lowestPitch.midi:
                    highestPitch = sorted_pitches[-1]
                    # this is an ugly hack to overcome the cases where part id is lost in chordify
                    if (highestPitch.midi-lowestPitch.midi) > 11 or lowestPitch.midi <= 72:
                        retVal = lowestPitch
        return retVal

    def compareBassWithPitch(self, pitchClass, mode='any', favor_by_part=False, HarmonicState=None):
        if HarmonicState is None:
            HarmonicState = self.CurrHarmonicState
        retVal = False
        bassPitch = []
        if mode in ['any', 'restless']:
            if not retVal:
                bassPitch = self.getBassFromChord(HarmonicState.Chord, favor_by_part=favor_by_part)
                if bassPitch:
                    retVal = bassPitch.pitchClass==pitchClass
        if mode in ['any', 'orig']:
            if not bassPitch:
                bassPitch = self.getBassFromChord(HarmonicState.ChordWithBassRests, favor_by_part=favor_by_part)
                if bassPitch:
                    retVal = bassPitch.pitchClass==pitchClass
        return retVal

    def isSopranoOnDegree(self, deg):
        retVal = 0
        if not isinstance(deg, list):
            deg = [deg]
        for curr_deg in deg:
            target_pitch_class = self.CurrHarmonicState.Key.pitchFromDegree(curr_deg).pitchClass
            if self.CurrHarmonicState.Chord.pitches[-1].pitchClass == target_pitch_class:
                retVal = 1
                #check for voice crossing of other voices
            elif self.CurrHarmonicState.Chord.sortFrequencyAscending().pitches[-1].pitchClass == target_pitch_class:
                retVal = 1
                #in multipart scores, check for explicit soprano
            elif self.NumParts > 2:
                sopranoPartPitch = self.getSopranoPartPitch(self.CurrHarmonicState)
                if sopranoPartPitch and sopranoPartPitch.pitchClass == target_pitch_class:
                    retVal = 1
            if retVal:
                break
        return retVal

    def getSopranoPartPitch(self, HarmonicState):
        retVal  = 0
        if not HarmonicState.Chord.isRest:
            for p in HarmonicState.Chord.pitches:
                if 'MySoprano' in p.groups:
                    retVal = p
                    break
        return retVal

    def getSopranoPitch(self, HarmonicState):
        retVal = 0
        if not HarmonicState.Chord.isRest:
            retVal = HarmonicState.Chord.sortFrequencyAscending().pitches[-1]
        return retVal

    def getCurrSopranoPitch(self):
        return self.getSopranoPitch(self.CurrHarmonicState)

    def getPrevSopranoPitch(self):
        return self.getSopranoPitch(self.PrevHarmonicState)

    def isSopranoAtSemitoneFromDegree(self, n_semitones, deg):
        retVal = 0
        if not isinstance(deg, list):
            deg = [deg]
        for curr_deg in deg:
            if abs(self.CurrHarmonicState.Chord.pitches[-1].pitchClass - self.CurrHarmonicState.Key.pitchFromDegree(curr_deg).pitchClass) == n_semitones:
                 retVal = 1
                 #check for voice crossing of other voices
            elif abs(self.CurrHarmonicState.Chord.sortFrequencyAscending().pitches[-1].pitchClass - self.CurrHarmonicState.Key.pitchFromDegree(curr_deg).pitchClass) == n_semitones:
                 retVal = 1
            else:
                for p in self.CurrHarmonicState.Chord.pitches:
                    if abs(p.pitchClass - self.CurrHarmonicState.Key.pitchFromDegree(curr_deg).pitchClass) == n_semitones and 'MySoprano' in p.groups:
                        retVal = 1
                        break
            if retVal:
                break
        return retVal

    def isPartInChord(self, chord, partString):
        retVal = 0
        for p in chord.pitches:
            if partString in p.groups:
                retVal = 1
                break
        return retVal

    def getPartPitchFromChord(self, chord, partString):
        retPitch = 0
        for p in chord.pitches:
            if partString in p.groups:
                retPitch = p
                break
        return retPitch

    def isDominantBass(self, favor_by_part=False, HarmonicState=None):
        if HarmonicState is None:
            HarmonicState = self.CurrHarmonicState
        dominantPitchClass = HarmonicState.Key.pitchFromDegree(5).pitchClass
        # chordify may miss the correct bass, check all notes by frequency ascending
        retVal = self.compareBassWithPitch(dominantPitchClass, mode='any', favor_by_part=favor_by_part, HarmonicState=HarmonicState)
        return retVal

    def isSubDominantBass(self, favor_by_part=False):
        subDominantPitchClass = self.CurrHarmonicState.Key.pitchFromDegree(4).pitchClass
        # chordify may miss the correct bass, check all notes by frequency ascending
        retVal = self.compareBassWithPitch(subDominantPitchClass, mode='any', favor_by_part=favor_by_part)
        return retVal

    def isMediantBass(self, favor_by_part=False):
        dominantPitchClass = self.CurrHarmonicState.Key.pitchFromDegree(3).pitchClass
        # chordify may miss the correct bass, check all notes by frequency ascending
        retVal = self.compareBassWithPitch(dominantPitchClass, mode='any', favor_by_part=favor_by_part)
        return retVal

    def isTonicBass(self):
        tonicPitchClass = self.CurrHarmonicState.Key.pitchFromDegree(1).pitchClass
        retVal = self.compareBassWithPitch(tonicPitchClass, mode='any', favor_by_part=True)
        return retVal

    def harmonyHasThird(self):
        return not (self.CurrHarmonicState.Chord.third is None)

    def isMatchingQuality(self):
        return not self.harmonyHasThird() or\
               (self.CurrHarmonicState.RomanNumeral.quality == self.CurrHarmonicState.Key.mode)

    def harmonyHasSeventh(self, HarmonicState):
        return self.harmonyContainsPitchDegree(HarmonicState=HarmonicState, degree=4)
        #return not (HarmonicState.Chord.seventh is None)

    def harmonyHasTonic(self,  HarmonicState):
        return self.harmonyContainsPitchDegree(HarmonicState=HarmonicState, degree=1)

    def harmonyHasMediant(self, HarmonicState):
        return self.harmonyContainsPitchDegree(HarmonicState=HarmonicState, degree=3)

    # TBD - should obsolete this function
    def verifyNotI64(self, HarmonicState):
        return not (self.harmonyHasTonic(HarmonicState) or self.harmonyHasMediant(HarmonicState)) or self.harmonyHasSeventh(HarmonicState)

    def isI64(self, HarmonicState):
        return self.harmonyContainsOnlyPitchDegrees(HarmonicState=HarmonicState, degrees=[1, 3, 5]) and\
               (self.harmonyHasTonic(HarmonicState) or self.harmonyHasMediant(HarmonicState)) and\
               self.isDominantBass(HarmonicState=HarmonicState)

    def harmonyContainsOnlyPitchDegrees(self, HarmonicState, degrees):
        pitch_classes = [HarmonicState.Key.pitchFromDegree(degree).pitchClass for degree in degrees]
        return self.harmonyContainsOnlyPitchClasses(HarmonicState=HarmonicState, pitch_classes=pitch_classes)

    def harmonyContainsOnlyPitchClasses(self, HarmonicState, pitch_classes):
        retVal = True
        if not HarmonicState.ChordWithBassRests.isRest:
            for p in HarmonicState.ChordWithBassRests.pitches:
                if p.pitchClass not in pitch_classes:
                    retVal = False
                    break
        return retVal

    def isDominantHarmony(self, HarmonicState):
        dominant_harmony_degrees = [5, 2, 4] # first without leading tone
        pitch_classes = [HarmonicState.Key.pitchFromDegree(degree).pitchClass for degree in dominant_harmony_degrees]
        leading_tone_pitch_class = HarmonicState.Key.pitchFromDegree(7).pitchClass
        if HarmonicState.Key.mode == 'minor':
            leading_tone_pitch_class = (leading_tone_pitch_class + 1) % 12 # raise the leading tone
        pitch_classes.append(leading_tone_pitch_class)
        return self.harmonyContainsOnlyPitchClasses(HarmonicState=HarmonicState, pitch_classes=pitch_classes)

    def harmonyContainsPitchDegree(self, HarmonicState, degree):
        retVal = False
        if not HarmonicState.ChordWithBassRests.isRest:
            PitchClass = HarmonicState.Key.pitchFromDegree(degree).pitchClass
            retVal = PitchClass in [p.pitchClass for p in HarmonicState.ChordWithBassRests.pitches]
        return retVal

    def isLeadingToneSoprano(self):
        interval_to_tonic = self.CurrHarmonicState.Chord.sortFrequencyAscending().pitches[-1].pitchClass - self.CurrHarmonicState.Key.tonic.pitchClass
        return interval_to_tonic % 12 == 11

    def isSopranoOnSemitonesFromTonic(self, intervalsFromTonic):
        interval_to_tonic = self.CurrHarmonicState.Chord.sortFrequencyAscending().pitches[-1].pitchClass - self.CurrHarmonicState.Key.tonic.pitchClass
        return interval_to_tonic % 12 in intervalsFromTonic

    def isLeadingToneBass(self):
        retVal = False
        bassPitch = self.getBassFromChord(self.CurrHarmonicState.Chord, favor_by_part=True)
        if not bassPitch:
            bassPitch = self.getBassFromChord(self.CurrHarmonicState.ChordWithBassRests, favor_by_part=True)
        if bassPitch:
            interval_to_tonic = bassPitch.pitchClass - self.CurrHarmonicState.Key.tonic.pitchClass
            retVal = interval_to_tonic % 12 == 11
        return retVal

    def isRootedHarmony(self):
        return self.compareBassWithPitch(self.CurrHarmonicState.Chord.root().pitchClass, mode='any')

    def isSecondaryDominantLeadingTone(self):
        retVal= 0
        Fourth = self.CurrHarmonicState.Key.pitchFromDegree(4)
        RaisedFourth = pitch.Pitch(midi=Fourth.midi+1)
        pSecondaryLeadingTone = RaisedFourth.pitchClass
        if not self.CurrHarmonicState.Chord.isRest:
            for p in self.CurrHarmonicState.Chord.pitches:
                if p.pitchClass == pSecondaryLeadingTone:
                    retVal = 1
                    break
        return retVal

    def isSecondaryDominantUpperLeadingTone(self):
        retVal= 0
        if not self.CurrHarmonicState.Chord.isRest:
            Fifth = self.CurrHarmonicState.Key.pitchFromDegree(5)
            RaisedFifth = pitch.Pitch(midi=Fifth.midi+1)
            pSecondaryDominantUpperLeadingTone = RaisedFifth.pitchClass
            for p in self.CurrHarmonicState.Chord.pitches:
                if p.pitchClass == pSecondaryDominantUpperLeadingTone:
                    retVal = 1
                    break
        return retVal

    def isDominantEmbellishment(self):
        retVal = 0
        VpitchClass = self.CurrHarmonicState.Key.pitchFromDegree(5).pitchClass
        if self.CurrHarmonicState.Key.mode == 'major':
            # Vb,V#,VIIb,IIb,II#
            embellishment_offsets = [-1, 1, 3, 6, 8]
        else:
            # Vb,V#,VIIb,IIb
            embellishment_offsets = [-1, 1, 3, 6]
        embellishments = [(VpitchClass + eo) % 12 for eo in embellishment_offsets]
        if not self.CurrHarmonicState.Chord.isRest:
            for p in self.CurrHarmonicState.Chord.pitches:
                if p.pitchClass in embellishments:
                    retVal = 1
                    break
        return retVal

    def isVMajorExclusive(self, HarmonicState):
        retVal = 1
        VpitchClass = self.CurrHarmonicState.Key.pitchFromDegree(5).pitchClass
        # V,VII,II
        major_triad_offsets = [0, 4, 7]
        VchordPitchClass = [(VpitchClass + mto) % 12 for mto in major_triad_offsets]
        if not HarmonicState.Chord.isRest:
            for p in HarmonicState.Chord.pitches:
                if p.pitchClass not in VchordPitchClass:
                    retVal = 0
                    break
        return retVal

    def verifyHCHarmony(self, HarmonicState):
        return self.isVMajorExclusive(HarmonicState)

    def isUnison(self):
        retVal = 1
        Pitch0 = self.CurrHarmonicState.Chord.pitches[0].pitchClass
        for p in self.CurrHarmonicState.Chord.pitches:
            if p.pitchClass != Pitch0:
                retVal = 0
                break
        return retVal

    def numTonics(self):
        nTonics = 0
        pTonic = self.CurrHarmonicState.Key.pitchFromDegree(1).pitchClass
        for p in self.CurrHarmonicState.Chord.pitches:
            if p.pitchClass == pTonic:
                nTonics = nTonics + 1
        print("nTonics:",nTonics)
        return nTonics

    def tryGetBeatStrength(self):
        retVal = 0
        try:
            retVal = self.CurrHarmonicState.Chord.beatStrength
        except:
            retVal = 0
            print('error: could not get beat strength. Returning 0')
        return retVal

    def verifyPACGrouping(self, HarmonicState):
        return self.verifyGrouping(HarmonicState=HarmonicState, start_stop='any') or self.tryGetBeatStrength()==1.0

    def verifyIACGrouping(self, HarmonicState):
        return self.verifyGrouping(HarmonicState=HarmonicState, start_stop='stop')

    def verifyHCGrouping(self, HarmonicState, start_stop='stop'):
        return self.verifyGrouping(HarmonicState=HarmonicState, start_stop=start_stop) and self.tryGetBeatStrength()>=0.25

    def verifyCadenceVoiceLeading(self, prev_soprano_pitch, curr_soprano_pitch, cadence_type='PAC'):
        retVal = True
        if curr_soprano_pitch and prev_soprano_pitch:
            midi_diff = curr_soprano_pitch.midi - prev_soprano_pitch.midi
            allowed_interval_table = {'PAC': np.array([0, 1, -2, -3, -4, 5, -7]), # anticipated tonic, minor second up, major second down, major/minor third down, fourth up, fifth down
                                      'IAC': np.array([0, 1, 2, 3, 4, -1, -2, -3, -4, 5]),  # anticipated third or fifth, major/minor second up, major/minor second down, major/minor third up and down, fourth up
                                      'HC': np.array([0, 1, 2, -1, -2, -5, -7])}
            allowed_intervals = allowed_interval_table[cadence_type]
            allowed_intervals_full = np.concatenate((allowed_intervals, allowed_intervals-12, allowed_intervals+12)) # check also an octave down and up, Mozart K533-1, Haydn17_3_4
            retVal = midi_diff in allowed_intervals_full
        return retVal

    def verifySopranoVoiceLeading(self, cadence_type='PAC'):
        curr_soprano_pitch = self.getSopranoPitch(self.CurrHarmonicState)
        prev_soprano_pitch = self.getSopranoPitch(self.PrevHarmonicState)
        retVal = self.verifyCadenceVoiceLeading(prev_soprano_pitch, curr_soprano_pitch, cadence_type=cadence_type)
        if not retVal:
            curr_soprano_pitch = self.getSopranoPartPitch(self.CurrHarmonicState)
            prev_soprano_pitch = self.getSopranoPartPitch(self.PrevHarmonicState)
            retVal = self.verifyCadenceVoiceLeading(prev_soprano_pitch, curr_soprano_pitch, cadence_type=cadence_type)
        return retVal

    def verifyGrouping(self, HarmonicState, start_stop='any'):
        search_sign = {'start': ['{'], 'stop': ['}'], 'any': ['{','}']}
        for real_part in HarmonicState.RealNotes:
            for note in real_part:
                for sign in search_sign[start_stop]:
                    if sign in note.groups:
                        return True
        return False

    def verifySopranoGrouping(self, HarmonicState, start_stop='any'):
        search_sign = {'start': ['{'], 'stop': ['}'], 'any': ['{', '}']}
        retVal  = False
        highest_pitch_midi = 0
        for real_part in HarmonicState.RealNotes:
            for note in real_part:
                if note.isChord:
                    curr_note = note.sortFrequencyAscending()[-1]
                else:
                    curr_note = note
                if curr_note.pitch.midi >= highest_pitch_midi:
                    highest_pitch_midi = curr_note.pitch.midi
                    for sign in search_sign[start_stop]:
                        retVal = sign in note.groups
        return retVal

    def isBassPartRest(self):
        retVal = True
        # serach for explicit rest in bass part
        bass_part = self.CurrHarmonicState.RealNotes[-1]
        if len(bass_part) == 0:
            retVal = False
        else:
            for noteOrRest in bass_part:
                if not noteOrRest.isRest:
                    retVal = False
                    break
        return retVal


    def updateCadentialState(self):

        curr_state = self.CurrCadentialState #set to temp variable for clean code

        # ====on key change, return to Cadence expected, not idle since the key change itself creates the initial tension
        if self.KeyChangeOneShot==1:
            curr_state = CDCadentialStates.CadExpected

        # ===============================================
        # ====idle state, wait for IV or II6 or I6=======
        # ===============================================
        if curr_state == CDCadentialStates.Idle:
            if (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio):
                # do nothing
                curr_state = curr_state
            elif self.CurrHarmonicState.ChordWithBassRests.isRest and not self.PrevHarmonicState.Chord.isRest:
                if self.tryGetBeatStrength() < 1.0:
                    #if sudden rest and not on strongest beat, check if previous harmony was dominant
                    if isinstance(self.PrevHarmonicState.Key, key.Key):
                        dominantPitchClass = self.PrevHarmonicState.Key.pitchFromDegree(5).pitchClass
                        bassPitch = self.getBassFromChord(self.PrevHarmonicState.Chord)
                        if bassPitch and bassPitch.pitchClass == dominantPitchClass and not self.isI64(HarmonicState=self.PrevHarmonicState) and self.verifySopranoVoiceLeading(cadence_type='HC'):
                            curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                else:
                    curr_state = curr_state
            elif self.isDominantEmbellishment():# this can create false HCs, need to consider non-cadential chromatic situations
                curr_state = CDCadentialStates.HCArrivalExpected
            elif self.CurrHarmonicState.ChordWithBassRests.isRest:
                curr_state = curr_state
            elif self.compareBassWithPitch(self.CurrHarmonicState.Key.pitchFromDegree(4).pitchClass, mode='any'):
                #== bass in 4th degree - IV or II6 go to expecting cadence
                curr_state = CDCadentialStates.CadExpected
            elif self.compareBassWithPitch(self.CurrHarmonicState.Key.pitchFromDegree(6).pitchClass, mode='any'):
                #== bass in 6th degree - VI or IV6 go to expecting cadence
                curr_state = CDCadentialStates.CadExpected
            elif self.tryGetBeatStrength() >= 0.5 and self.compareBassWithPitch(self.CurrHarmonicState.Key.pitchFromDegree(3).pitchClass, mode='any'):
                #== bass in 3rd degree - iii or I6 on strong beat go to expecting cadence
                curr_state = CDCadentialStates.CadExpected

        # ===================================================
        # ====expecting cadence, wait for V Dominant=========
        # ===================================================
        elif curr_state == CDCadentialStates.CadExpected or curr_state == CDCadentialStates.CadAvoided or curr_state == CDCadentialStates.IACArrival:
            # only stay in CadAvoided once (currently for display purposes)
            if curr_state == CDCadentialStates.CadAvoided:
                curr_state = CDCadentialStates.CadExpected
            elif curr_state == CDCadentialStates.IACArrival and not (self.isTonicBass() and self.isSopranoOnDegree(1)):
                curr_state = CDCadentialStates.CadExpected
            elif curr_state == CDCadentialStates.IACArrival and self.tryGetBeatStrength() == 1:
                curr_state = CDCadentialStates.CadExpected
            # now start the real state logic
            if curr_state == CDCadentialStates.IACArrival and (self.isTonicBass() and self.isSopranoOnDegree(1)):
                self.RevertLastIAC = True
                curr_state = self.setCadenceOrPostCadence(CDCadentialStates.PACArrival, curr_state, verify_beat_strength=False)
            elif self.isDominantEmbellishment():
                curr_state = CDCadentialStates.HCArrivalExpected
            elif self.CurrHarmonicState.ChordWithBassRests.isRest:
                curr_state = curr_state
            #ignoring arppegios
            elif (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio):
                #if alberti or arpeggio bass unless explicit V chord
                if self.isDominantBass(HarmonicState=self.LastNonAlbertiHarmonicState) and self.harmonyHasSeventh(self.CurrHarmonicState):
                    curr_state = CDCadentialStates.CadInevitable
                else:
                    curr_state = curr_state
            elif self.isDominantBass() and self.harmonyHasSeventh(self.CurrHarmonicState):
                curr_state = CDCadentialStates.CadInevitable
            elif self.isTonicBass():
                curr_state = CDCadentialStates.HCArrivalExpected
            elif self.isDominantBass(favor_by_part=True):
                if not self.isI64(self.CurrHarmonicState):
                    if self.verifySopranoVoiceLeading(cadence_type='HC') and \
                        ((self.tryGetBeatStrength()==1 and self.verifyHCGrouping(HarmonicState=self.CurrHarmonicState)) or
                         self.verifyHCGrouping(HarmonicState=self.CurrHarmonicState, start_stop='stop')) and\
                            self.verifyHCHarmony(HarmonicState=self.CurrHarmonicState):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                    else:
                        curr_state = CDCadentialStates.CadInevitable
                else:
                    curr_state = CDCadentialStates.HCAppoggExpected

        # ========================================================
        # ====inevitable cadence (PAC or IAC), wait for Is========
        # ========================================================
        elif curr_state==CDCadentialStates.CadInevitable or curr_state==CDCadentialStates.IACArrivalExpected:
            #on dominant and then a complete rest --> HC with constraints
            if self.CurrHarmonicState.ChordWithBassRests.isRest:
                if self.verifySopranoVoiceLeading(cadence_type='HC') and\
                        self.verifyHCGrouping(HarmonicState=self.PrevHarmonicState) and\
                        self.verifyHCHarmony(HarmonicState=self.PrevHarmonicState):
                    curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                else:
                    curr_state = curr_state
            # on dominant and then bass rest --> HC with constraints
            elif self.isBassPartRest() and\
                    self.isDominantBass(favor_by_part=True) and\
                    self.verifyHCGrouping(HarmonicState=self.CurrHarmonicState) and\
                    self.verifyHCHarmony(HarmonicState=self.CurrHarmonicState):
                curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
            elif self.isDominantEmbellishment():
                curr_state = CDCadentialStates.HCArrivalExpected
            elif (self.isI64(HarmonicState=self.CurrHarmonicState) and not self.isBassPartRest()):
                curr_state = CDCadentialStates.HCAppoggExpected
            elif (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio):
                #if alberti and weak beat do nothing
                curr_state = curr_state
            else:
                # for the case where in IAC expected and dominant appears again, back to Cad inevitable
                if curr_state==CDCadentialStates.IACArrivalExpected and self.isDominantBass(favor_by_part=True) and not self.isBassPartRest():
                    curr_state = CDCadentialStates.CadInevitable
                # meter - look for cadences on strong beat:
                elif self.tryGetBeatStrength() >= 0.25:#cadence can only occur on strong beats
                    # harmony  - chordal degree and bass analysis
                    # bass rest on strong beat --> HC
                    if self.isDominantBass(favor_by_part=True) and\
                            self.isBassPartRest() and\
                            self.tryGetBeatStrength() < 1.0 and\
                            self.verifyHCGrouping(self.CurrHarmonicState) and\
                            self.verifyHCHarmony(self.CurrHarmonicState) and\
                            self.verifySopranoVoiceLeading(cadence_type='HC'):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                    elif self.isTonicBass():
                        # harmony  - chordal inversion
                        if (self.CurrHarmonicState.ChordInversion == CDHarmonicChordInversions.Root.value or self.isRootedHarmony()) and\
                                self.verifyPACGrouping(HarmonicState=self.CurrHarmonicState):
                            # ==I after V after IV or II6, cadential arrival
                            # melody  - soprano degree
                            if self.isSopranoOnDegree(1) and self.verifySopranoVoiceLeading(cadence_type='PAC'):
                                if curr_state==CDCadentialStates.CadInevitable:
                                    curr_state = self.setCadenceOrPostCadence(CDCadentialStates.PACArrival, curr_state)
                                elif self.verifySopranoVoiceLeading(cadence_type='IAC') and self.verifyIACGrouping(HarmonicState=self.CurrHarmonicState):
                                    curr_state = self.setCadenceOrPostCadence(CDCadentialStates.IACArrival, curr_state)
                            #soprano not on 1, either IAC or appoggiatura
                            elif self.isSopranoOnDegree([3, 5]) and self.verifySopranoVoiceLeading(cadence_type='IAC') and self.verifyIACGrouping(HarmonicState=self.CurrHarmonicState):
                                curr_state = self.setCadenceOrPostCadence(CDCadentialStates.IACArrival, curr_state)
                            # expecting appoggiatura on strong beats
                            elif self.tryGetBeatStrength() >= 0.5 and self.verifyPACGrouping(self.CurrHarmonicState):
                                curr_state = self.checkAppoggiaturaOrAvoidance('PAC' if curr_state==CDCadentialStates.CadInevitable else 'IAC')
                            # appogiatura can also not be detected as I
                        elif self.tryGetBeatStrength() >= 0.5 and self.verifyPACGrouping(self.CurrHarmonicState):
                            # expecting appogiaturas only on strongest beats (TBD - this might be overfit to haydn)
                            curr_state = self.checkAppoggiaturaOrAvoidance('PAC' if curr_state==CDCadentialStates.CadInevitable else 'IAC')
                    # on strong beat: going from V to anything other than V or I is avoiding the cadence or bass appoggiatura
                    elif self.tryGetBeatStrength() >= 0.5 and self.verifyPACGrouping(self.CurrHarmonicState) and self.verifySopranoVoiceLeading(cadence_type='PAC'):
                        # expecting bass appogiaturas only on strongest beats
                        curr_state = self.checkBassAppoggiatura(curr_state)
                    elif not self.isDominantBass():
                        if self.tryGetBeatStrength() >= 0.5:
                            curr_state = CDCadentialStates.CadAvoided
                        else:
                            curr_state = CDCadentialStates.CadExpected
                # on weaker beat (but not completely weak) leave this state if not dominant bass:
                elif self.tryGetBeatStrength() >= 0.1 and not self.isDominantBass() and not self.isLeadingToneBass():
                    # if we left dominant but not to I then cadence avoided, but could be HC so wait for V again - TBD, perhaps this avoidance goes further back
                    curr_state = CDCadentialStates.HCArrivalExpected

        # =============================================================================================
        # ====HC expected, V on strong beat = HC, V on weak beat, return to cadence inevitable ========
        # =============================================================================================

        elif curr_state == CDCadentialStates.HCArrivalExpected:

            if self.isDominantBass(favor_by_part=True) and self.harmonyHasSeventh(self.CurrHarmonicState):
                curr_state = CDCadentialStates.CadInevitable
            elif self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio:
                if self.isDominantBass(favor_by_part=True, HarmonicState=self.LastNonAlbertiHarmonicState) and\
                        self.harmonyHasSeventh(self.CurrHarmonicState) and\
                        not self.isDominantEmbellishment():
                    curr_state = CDCadentialStates.CadInevitable
                else:
                    curr_state = curr_state
            elif self.tryGetBeatStrength() < 0.5 and\
                    self.isDominantBass(favor_by_part=True) and\
                    self.isDominantHarmony(HarmonicState=self.CurrHarmonicState):#weak beat returning to dominant, return to cad inevitable
                curr_state = CDCadentialStates.CadInevitable
            elif self.CurrHarmonicState.ChordWithBassRests.isRest:
                curr_state = CDCadentialStates.CadExpected
            elif self.tryGetBeatStrength() >= 0.5: #strong beats
                if self.isDominantBass(favor_by_part=True) and\
                        (self.verifyHCGrouping(self.CurrHarmonicState, start_stop='any') or self.tryGetBeatStrength()==1) and\
                        not self.isI64(HarmonicState=self.CurrHarmonicState) and\
                        self.verifyHCHarmony(HarmonicState=self.CurrHarmonicState) and\
                        self.verifySopranoVoiceLeading(cadence_type='HC'):
                    curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                # appoggiatura on strongest beat, or if V64
                elif self.isDominantBass(favor_by_part=True) and\
                        (self.isI64(HarmonicState=self.CurrHarmonicState) and not self.isBassPartRest()) and\
                        self.verifyHCGrouping(self.CurrHarmonicState, start_stop='any'):
                    curr_state = CDCadentialStates.HCAppoggExpected
                elif self.isDominantBass(favor_by_part=True) and\
                        ((self.harmonyHasSeventh(self.CurrHarmonicState) and not self.isDominantEmbellishment()) or\
                         not (self.verifyHCGrouping(self.CurrHarmonicState, start_stop='any') or self.tryGetBeatStrength()==1)): #V7 or V without grouping return to cad inevitable
                    curr_state = CDCadentialStates.CadInevitable
                elif self.isDominantBass(favor_by_part=True) and self.tryGetBeatStrength() == 1.0:#appoggiatura on strongest beat, or if V64
                    curr_state = CDCadentialStates.HCAppoggExpected
                elif not (self.isDominantBass() or self.isDominantEmbellishment()): #strong beat and not dominant or embellishment, cadence avoided
                    curr_state = CDCadentialStates.CadAvoided
                elif self.isDominantBass(favor_by_part=True) and self.verifyHCGrouping(HarmonicState=self.CurrHarmonicState, start_stop='any') and not self.harmonyHasSeventh(self.CurrHarmonicState):
                    curr_state = CDCadentialStates.HCAppoggExpected


        elif curr_state == CDCadentialStates.PACAppoggExpected or curr_state == CDCadentialStates.BassAppoggExpected:

            if self.CurrHarmonicState.ChordWithBassRests.isRest or (self.tryGetBeatStrength() == 1): # and not self.isTonicBass()): # rest or new measure not on tonic, exit appoggiatura:
                if self.isDominantBass():
                    if (self.verifyHCGrouping(self.CurrHarmonicState, start_stop='any') or self.tryGetBeatStrength()==1) and\
                            self.verifyHCHarmony(HarmonicState=self.CurrHarmonicState) and\
                            self.verifySopranoVoiceLeading(cadence_type='HC'):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                    elif self.isI64(self.CurrHarmonicState):  # I64 can be HC appoggiaturra on dominant bass
                        curr_state = CDCadentialStates.HCAppoggExpected
                    else:
                        curr_state = CDCadentialStates.CadInevitable
                elif self.isTonicBass() and\
                        self.CurrHarmonicState.Chord.isConsonant() and\
                        not self.harmonyContainsPitchDegree(HarmonicState=self.CurrHarmonicState, degree=6):
                    if self.isSopranoOnDegree(1) and self.verifySopranoVoiceLeading(cadence_type='PAC') and\
                            (not self.IsChallenger or self.verifyPACGrouping(self.CurrHarmonicState)):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.PACArrival, curr_state)
                    elif self.isSopranoOnDegree([3, 5]) and self.verifySopranoVoiceLeading(cadence_type='IAC') and\
                            self.verifyIACGrouping(HarmonicState=self.CurrHarmonicState):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.IACArrival, curr_state)
                    else:
                        curr_state = CDCadentialStates.CadExpected
                else:
                    curr_state = CDCadentialStates.CadExpected
            else:
            # ==appoggiatura, check bass still on key and if soprano is root then PAC otherwise IAC
            # ==for bass appoggiatura check  that soprano is still on tonic and that bass is also.
            # ==grouping is not required becuase it was verified on the appoggiatura entrance
            # ==in case of alberti or arppeggio the original bass on entrance remains, so tonic bass verification is not required.
                if curr_state == CDCadentialStates.BassAppoggExpected and (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio):
                    curr_state = curr_state
                elif self.isTonicBass() or\
                        (curr_state == CDCadentialStates.PACAppoggExpected and (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio)) and\
                        self.CurrHarmonicState.Chord.isConsonant() and\
                        not self.harmonyContainsPitchDegree(HarmonicState=self.CurrHarmonicState, degree=6):
                    if self.isSopranoOnDegree(1) and self.verifySopranoVoiceLeading(cadence_type='PAC') and\
                            (not self.IsChallenger or self.verifyPACGrouping(self.CurrHarmonicState)):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.PACArrival, curr_state)
                    elif self.isSopranoOnDegree(3) and self.verifySopranoVoiceLeading(cadence_type='IAC') and \
                            self.verifyIACGrouping(HarmonicState=self.CurrHarmonicState):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.IACArrival, curr_state)
                elif self.isDominantBass(favor_by_part=True) and\
                        (self.verifyHCGrouping(self.CurrHarmonicState, start_stop='any') or self.tryGetBeatStrength()==1) and\
                        self.verifyHCHarmony(HarmonicState=self.CurrHarmonicState) and\
                        self.verifySopranoVoiceLeading(cadence_type='HC'):
                    curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                elif not (self.isSopranoOnDegree(1) or self.isDominantBass() or self.isMediantBass()):#continue the appogiattura if dominant or mediant bass
                    curr_state = CDCadentialStates.CadAvoided
                elif (self.IsChallenger and self.verifyPACGrouping(self.CurrHarmonicState)):
                    curr_state = CDCadentialStates.CadAvoided


        elif curr_state == CDCadentialStates.IACAppoggExpected:
            # ==appoggiatura, check bass still on key and if soprano is root then PAC otherwise IAC
            if self.isDominantEmbellishment() or (self.isI64(HarmonicState=self.CurrHarmonicState) and not self.isBassPartRest()):
                curr_state = CDCadentialStates.HCArrivalExpected
            elif self.CurrHarmonicState.ChordWithBassRests.isRest or self.tryGetBeatStrength() == 1: # rest or new measure, exit appoggiatura
                curr_state = CDCadentialStates.CadExpected
            else:
                if self.isTonicBass():
                    if ((self.isSopranoOnDegree(1) and self.verifySopranoVoiceLeading(cadence_type='PAC')) or
                        (self.isSopranoOnDegree([3, 5]) and self.verifySopranoVoiceLeading(cadence_type='IAC'))) and\
                            (not self.IsChallenger or self.verifyPACGrouping(self.CurrHarmonicState)):
                        if self.CurrHarmonicState.Chord.isConsonant() and not self.harmonyContainsPitchDegree(HarmonicState=self.CurrHarmonicState, degree=6):
                            curr_state = self.setCadenceOrPostCadence(CDCadentialStates.IACArrival, curr_state)
                elif self.isDominantBass(): # dominant bass, go to PAC appogiatura
                    curr_state = self.checkAppoggiaturaOrAvoidance('PAC')
                elif not (self.isSopranoOnDegree(1) or self.isMediantBass()):#continue the appogiattura if dominant or mediant bass
                    curr_state = CDCadentialStates.CadAvoided


        elif curr_state == CDCadentialStates.HCAppoggExpected:
            # ==HC with appoggiatura, don't exit as long as bass is dominant
            if self.CurrHarmonicState.ChordWithBassRests.isRest:
                curr_state = CDCadentialStates.CadExpected
            else:
                if (self.isI64(HarmonicState=self.CurrHarmonicState) and not self.isBassPartRest()) or self.isDominantEmbellishment(): # stay in HC appogg as long as I64 or dominant embellishment
                    curr_state = curr_state
                elif self.isDominantBass(favor_by_part=True) and\
                        self.verifyHCHarmony(self.CurrHarmonicState) and\
                        self.verifySopranoVoiceLeading(cadence_type='HC'):
                    curr_state = self.setCadenceOrPostCadence(CDCadentialStates.HCArrival, curr_state)
                #elif self.isDominantBass(favor_by_part=True) and\
                #        (self.harmonyHasSeventh(self.CurrHarmonicState) or self.tryGetBeatStrength() >= 0.5): #V7 or strong beat, return to cad inevitable
                #    curr_state = CDCadentialStates.CadInevitable
                #elif self.harmonyContainsPitchDegree(HarmonicState=self.CurrHarmonicState, degree=4): #also V7
                #    curr_state = CDCadentialStates.CadInevitable
                elif (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio):
                    # V7 in alberti
                    if self.isDominantBass(favor_by_part=True, HarmonicState=self.LastNonAlbertiHarmonicState) and\
                            self.harmonyHasSeventh(self.CurrHarmonicState) and\
                            not self.isDominantEmbellishment():
                        curr_state = CDCadentialStates.CadInevitable
                    else:
                        curr_state = curr_state
                elif self.isTonicBass() and self.tryGetBeatStrength() >= 0.5 and\
                        self.verifyPACGrouping(HarmonicState=self.CurrHarmonicState) and\
                        self.isDominantHarmony(HarmonicState=self.PrevHarmonicState):
                    if self.isSopranoOnDegree(1) and self.verifySopranoVoiceLeading(cadence_type='PAC'):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.PACArrival, curr_state)
                    elif self.isSopranoOnDegree([3, 5]) and self.verifySopranoVoiceLeading(cadence_type='IAC'):
                        curr_state = self.setCadenceOrPostCadence(CDCadentialStates.IACArrival, curr_state)
                    else:
                        curr_state = self.checkAppoggiaturaOrAvoidance('PAC')
                elif self.tryGetBeatStrength() >= 0.25 and self.isDominantBass(favor_by_part=True) and (self.harmonyHasSeventh(self.CurrHarmonicState) or self.tryGetBeatStrength() >= 0.5): #V7 or strong beat, return to cad inevitable
                    curr_state = CDCadentialStates.CadInevitable
                elif not self.isDominantBass():
                    curr_state = CDCadentialStates.CadAvoided

        elif curr_state == CDCadentialStates.PACArrival or curr_state == CDCadentialStates.PCCArrival:
            #conditions for exiting PAC state:
            #1. complete rest
            #2. new measure
            #3. bass not on tonic
            #4. no soprano (TBD - what about non voiced works such as piano?), or lowered soprano
            if self.CurrHarmonicState.ChordWithBassRests.isRest or\
                    self.tryGetBeatStrength()>0.5 or\
                    not self.isTonicBass() or\
                    not self.isPartInChord(self.CurrHarmonicState.Chord,'MySoprano') or\
                    self.getCurrSopranoPitch().midi < self.SopranoOfLastCadence.midi:
                self.exitPACState()
                self.updateCadentialState()
                return
            else:
                # soprano still on tonic
                if self.isSopranoOnDegree(1):
                    self.PACPending = False
                    self.exitPACState()
                    self.updateCadentialState()
                    return
                elif self.tryGetBeatStrength() >= 0.25 and\
                        (self.isSopranoOnDegree(3)) and\
                        (self.CurrHarmonicState.Chord.isTriad() or
                         self.CurrHarmonicState.Chord.isIncompleteMajorTriad() or
                         self.CurrHarmonicState.Chord.isIncompleteMinorTriad()):
                    currSopranoPitch = self.getCurrSopranoPitch()
                    lastSopranoPitch = self.SopranoOfLastCadence
                    # a higher third in soprano following a PAC causes it retroactively to become a IAC!
                    print(currSopranoPitch,lastSopranoPitch)
                    if currSopranoPitch != 0 and lastSopranoPitch != 0:
                        if currSopranoPitch.midi > lastSopranoPitch.midi:
                            # moving up to 3, retroactively weakens the PAC to a IAC
                            self.PACPending = True
                        else:
                            # moving down to 3, confirms the PAC, move to idle and call update state again
                            self.PACPending = False
                            self.exitPACState()
                            self.updateCadentialState()
                            return


        elif curr_state == CDCadentialStates.HCArrival or curr_state == CDCadentialStates.PCHArrival:
            if self.CurrHarmonicState.ChordWithBassRests.isRest or self.PrevHarmonicState.ChordWithBassRests.isRest:# a full rest confirms the cadence
                curr_state = CDCadentialStates.Idle if self.MinPostCadenceMeasures > 0 else CDCadentialStates.CadInevitable
            elif (self.CurrHarmonicState.Alberti or self.CurrHarmonicState.Arpeggio):
                # if alberti and weak beat verify still not V7
                if self.harmonyContainsPitchDegree(HarmonicState=self.CurrHarmonicState, degree=4) and\
                        self.isDominantBass(HarmonicState=self.LastNonAlbertiHarmonicState):
                    curr_state = CDCadentialStates.CadInevitable
                    self.RevertLastHC = True
                    self.PostHCMeasureCounter = self.MinPostCadenceMeasuresHC
                else: # stay in state for an other Alberti or arpeggio
                    curr_state = curr_state
            else:
                # == After HC we are still expecting a cadence, but we need to see how this does not create false positives
                if self.isDominantBass() and self.harmonyHasSeventh(self.CurrHarmonicState): #V7, return to cad inevitable (and cancel the previous HC?)
                    curr_state = CDCadentialStates.CadInevitable
                    #if self.tryGetBeatStrength() <= 0.25: # this causes HC misses - example Haydn op.55,2,2 mm.34
                    #    self.RevertLastHC = True
                    #    self.PostHCMeasureCounter = self.MinPostCadenceMeasuresHC
                #elif self.isSubDominantBass(): # V2, cancel previous HC
                #    self.RevertLastHC = True
                #    self.PostHCMeasureCounter = self.MinPostCadenceMeasuresHC
                #    curr_state = CDCadentialStates.CadAvoided
                # === very strict condition from HC to PAC
                elif (self.isTonicBass() or self.isSopranoOnDegree(1)) and\
                        (self.tryGetBeatStrength() > self.WeightOfLastCadence or self.tryGetBeatStrength() > 0.25) and\
                        self.verifyPACGrouping(HarmonicState=self.CurrHarmonicState) and\
                        self.verifySopranoVoiceLeading(cadence_type='PAC'):
                    if self.isTonicBass():
                        if self.isSopranoOnDegree(1):
                            curr_state = self.setCadenceOrPostCadence(CDCadentialStates.PACArrival, curr_state)
                        elif self.isSopranoOnDegree(3):
                            curr_state = self.setCadenceOrPostCadence(CDCadentialStates.IACArrival, curr_state)
                        else:
                            curr_state = self.checkAppoggiaturaOrAvoidance('PAC')
                    elif self.isDominantBass() and self.isI64(HarmonicState=self.CurrHarmonicState):
                        curr_state = CDCadentialStates.HCArrivalExpected
                    else:
                        curr_state = self.checkBassAppoggiatura(curr_state=curr_state)
                elif not self.isDominantBass():#bass has left domninant, go back to cad expected
                    curr_state = CDCadentialStates.CadExpected
                else:
                    curr_state = curr_state

        #====set output to state and then set cadential arrivals back to idle state
        self.CurrCadentialOutput = curr_state
        self.CurrCadentialState = curr_state
        #print(self.CurrCadentialState)
        #==must check for change for flow debugging
        self.checkStateChanged()

    def checkAppoggiaturaOrAvoidance(self, app_type='PAC'):
        if app_type == 'PAC' and \
                ((self.CurrHarmonicState.Key.mode == 'major' and self.isSopranoOnSemitonesFromTonic([11, 2, 4])) or
                 (self.CurrHarmonicState.Key.mode == 'minor' and self.isSopranoOnSemitonesFromTonic([10, 11, 2, 3]))):
            curr_state = CDCadentialStates.PACAppoggExpected
        elif app_type == 'IAC' and self.isSopranoOnSemitonesFromTonic([3, 5]) and self.verifyIACGrouping(HarmonicState=self.CurrHarmonicState):
            curr_state = CDCadentialStates.IACAppoggExpected
        elif app_type == 'HC' and \
                ((self.CurrHarmonicState.Key.mode == 'major' and self.isSopranoOnSemitonesFromTonic([0, 1, 4, 6, 9])) or
                 (self.CurrHarmonicState.Key.mode == 'minor' and self.isSopranoOnSemitonesFromTonic([0, 1, 3, 6, 8]))):
            curr_state = CDCadentialStates.HCAppoggExpected
        else:
            curr_state = CDCadentialStates.CadAvoided
        return curr_state

    def checkBassAppoggiatura(self, curr_state):
        if self.isDominantBass(favor_by_part=True) or self.isMediantBass(favor_by_part=True) or self.isBassPartRest():
            if (curr_state in [CDCadentialStates.CadInevitable, CDCadentialStates.HCArrival, CDCadentialStates.PCHArrival]) and self.isSopranoOnDegree(1):
                curr_state = CDCadentialStates.BassAppoggExpected
            elif curr_state == CDCadentialStates.IACArrivalExpected and (self.isSopranoOnDegree([1, 2, 3]) or self.isLeadingToneSoprano()):
                curr_state = CDCadentialStates.IACAppoggExpected
            elif curr_state == CDCadentialStates.IACArrivalExpected and (self.isSopranoAtSemitoneFromDegree(1, 3)):
                curr_state = CDCadentialStates.IACAppoggExpected
            elif self.isDominantBass(favor_by_part=True):
                curr_state = CDCadentialStates.CadInevitable
            else:
                curr_state = CDCadentialStates.CadAvoided
        else:
            curr_state = CDCadentialStates.CadAvoided
        return curr_state

    def exitPACState(self):
        if self.PACPending:
            self.CurrCadentialState = CDCadentialStates.CadExpected
            self.RevertLastPAC = True
            self.PostPACMeasureCounter = self.MinPostCadenceMeasures #set the counter high to catch the next cadence
        else:
            # move to idle (confirming the PAC) and call update state again
            if self.MinPostCadenceMeasures > 0:
                self.CurrCadentialState = CDCadentialStates.Idle
            # if no minimum post cadence measures required, expect post cadential cadences
            else:
                self.CurrCadentialState = CDCadentialStates.CadExpected
            self.RevertLastPAC = False
        self.PACPending = False

    def updateCounters(self):
        #beat strenth==1 means new measure
        if self.tryGetBeatStrength() == 1:
            self.PostPACMeasureCounter += 1
            self.PostHCMeasureCounter += 1
            self.MeasureCounter += 1

    def setCadenceOrPostCadence(self, cadence=None, state=None, verify_beat_strength=True):
        if not verify_beat_strength or self.tryGetBeatStrength() >= 0.125:
            if self.MeasureCounter >= self.MinInitialMeasures[cadence]:
                if cadence == CDCadentialStates.PACArrival and self.PostPACMeasureCounter < self.MinPostCadenceMeasures:
                    state = CDCadentialStates.PCCArrival
                elif cadence == CDCadentialStates.HCArrival and self.PostHCMeasureCounter < self.MinPostCadenceMeasuresHC:
                    state = CDCadentialStates.PCHArrival
                else:
                    state = cadence
            else: # too early, return to cadence expected again
                state = CDCadentialStates.CadExpected
            if state == CDCadentialStates.HCArrival or state == CDCadentialStates.PCHArrival:
                self.WeightOfLastCadence = self.tryGetBeatStrength()
            self.SopranoOfLastCadence = self.getCurrSopranoPitch()
            self.HarmonicStateOfLastCadence = copy.deepcopy(self.CurrHarmonicState)
        else:
            state = state
        return state

    def getCadentialOutput(self):
        return self.CurrCadentialOutput

    def getCadentialOutputString(self):
        Lyric = str("")
        if self.ChangeFlagOneShot == 1:
            if self.getCadentialOutput() == CDCadentialStates.PACArrival:
                Lyric = str("PAC")
            elif self.getCadentialOutput() == CDCadentialStates.PCCArrival:
                Lyric = str("PCC")
            elif self.getCadentialOutput() == CDCadentialStates.IACArrival:
                Lyric = str("IAC")
            elif self.getCadentialOutput() == CDCadentialStates.HCArrival:
                Lyric = str("HC")
            elif self.getCadentialOutput() == CDCadentialStates.PCHArrival:
                Lyric = str("PCH")
            elif self.getCadentialOutput() == CDCadentialStates.CadAvoided:
                Lyric = str("CA")
            elif not (self.getCadentialOutput() == CDCadentialStates.Idle):
                Lyric = Lyric + self.TriggerString

        if self.KeyChangeOneShot == 1 or self.FirstKeyDetectionDone == 0 or self.CadentialKeyChange == 1:
            Lyric = Lyric + str("\nKey: ") + str(self.CurrHarmonicState.Key)
            self.FirstKeyDetectionDone = 1
            self.CadentialKeyChange = 0

        #debug arpeggios and alberti
        #if self.CurrHarmonicState.Alberti:
        #    Lyric = Lyric + ' Al'
        #elif self.CurrHarmonicState.Arpeggio:
        #    Lyric = Lyric + ' Ar'

        #Lyric = Lyric + " " + str(self.CurrHarmonicState.ChordDegree) + " " + str()
        return Lyric

    def getRevertLastPACAndReset(self):
        retVal = self.RevertLastPAC
        self.RevertLastPAC = False
        return retVal

    def getRevertLastHCAndReset(self):
        retVal = self.RevertLastHC
        self.RevertLastHC = False
        return retVal

    def getRevertLastIACAndReset(self):
        retVal = self.RevertLastIAC
        self.RevertLastIAC = False
        return retVal





from .CadenceDetector import *
import tqdm
from functools import partial
import os
import time
import multiprocessing as mp

def detect_cadences_in_file(full_path="", output_path=None):
    if not os.path.exists(full_path):
        print(f"file {full_path} not found")
        return
    CD = CadenceDetector()
    CD.loadFile(full_path)
    input_file_path, input_file_name = os.path.split(full_path)
    CD.setFileName(input_file_name)
    if output_path is None:
        output_path = input_file_path
    CD.setWritePath(output_path)
    CD.detectKeyPerMeasure()
    CD.writeKeyPerMeasureToFile()
    CD.readKeyPerMeasureFromFile()
    CD.detectCadences()
    try:
        CD.writeAnalyzedFile()
    except Exception as e:
        print('error: could not write file:', e)
    print(f"Analysis for {full_path} done.")
    return {'file': input_file_name, 'num_measures': CD.NumMeasures}


def detect_cadences_in_folder(input_dir="", file_ending=".xml", output_dir=None, do_parallel=True):
    fileList = sorted(os.listdir(input_dir))
    full_list = [os.path.join(input_dir, file) for file in fileList if file.endswith(file_ending)]
    start = time.time()
    num_measures_per_mov = []
    if output_dir is None:
        output_dir = input_dir
    if do_parallel:
        print("Parallel Processing On")
        print("Number of processors: ", mp.cpu_count())
        with mp.Pool() as pool:
            num_measures_per_mov = list(tqdm.tqdm(pool.imap_unordered(partial(detect_cadences_in_file, output_path=output_dir), full_list), total=len(full_list)))
        total_num_measures = sum([curr_tup['num_measures'] for curr_tup in num_measures_per_mov])
    else:
        print("Parallel Processing Off")
        total_num_measures = 0
        for file in tqdm.tqdm(full_list):
            num_measures_per_mov = detect_cadences_in_file(file, output_path=output_dir)
            if num_measures_per_mov is not None:
                total_num_measures = total_num_measures + num_measures_per_mov['num_measures']
    if num_measures_per_mov:
        for curr_mov in num_measures_per_mov:
            print(curr_mov)
        print("Total Num Measures:", total_num_measures)

    end = time.time()
    total_time = end - start
    print("Elapsed time", total_time/60, "minutes")
